# Evaluating the loop eq. Solves for different quantities
# when different masks are added.
# Uses Xn-1, Yn-1, Zn-1,  B_theta_n-2, Delta_n-1, B_psi_n-3, B_denom n-1,
# iota_coef (n-3)/2 or (n-2)/2, and B_alpha (n-1)/2 or (n-2)/2
# Must be evaluated with Z_coef_cp[n] = 0, p_perp_coef_cp[n] = 0
# B_psi_coef_cp[n-2] = 0, B_denom_coef_c[n] = 0 and B_theta_coef_cp[n] = 0
from math import floor, ceil
from aqsc.math_utilities import *
# from jax import jit
# from functools import partial
# @partial(jit, static_argnums=(0,))
def eval_loop(n, X_coef_cp, Y_coef_cp, Z_coef_cp, \
    B_theta_coef_cp, B_psi_coef_cp, B_alpha_coef, B_denom_coef_c, \
    p_perp_coef_cp, Delta_coef_cp, kap_p, dl_p, tau_p, iota_coef):
    def sum_arg_153(i288):
        # Child args for sum_arg_153
        return(X_coef_cp[i288]*diff(Y_coef_cp[n-i288],True,1))

    def sum_arg_152(i284):
        # Child args for sum_arg_152
        return(Y_coef_cp[i284]*diff(X_coef_cp[n-i284],True,1))

    def sum_arg_151(i256):
        # Child args for sum_arg_151
        return(i256*X_coef_cp[i256]*diff(Y_coef_cp[n-i256],True,1)+i256*diff(X_coef_cp[i256],True,1)*Y_coef_cp[n-i256])

    def sum_arg_150(i254):
        # Child args for sum_arg_150
        return((X_coef_cp[i254]*n-i254*X_coef_cp[i254])*diff(Y_coef_cp[n-i254],True,1)+(diff(X_coef_cp[i254],True,1)*n-i254*diff(X_coef_cp[i254],True,1))*Y_coef_cp[n-i254])

    def sum_arg_149(i288):
        # Child args for sum_arg_149
        return(X_coef_cp[i288]*diff(Y_coef_cp[n-i288],True,2)+diff(X_coef_cp[i288],True,1)*diff(Y_coef_cp[n-i288],True,1))

    def sum_arg_148(i288):
        # Child args for sum_arg_148
        return(X_coef_cp[i288]*diff(Y_coef_cp[n-i288],True,1,False,1)+diff(X_coef_cp[i288],False,1)*diff(Y_coef_cp[n-i288],True,1))

    def sum_arg_147(i288):
        # Child args for sum_arg_147
        return(X_coef_cp[i288]*diff(Y_coef_cp[n-i288],True,1))

    def sum_arg_146(i284):
        # Child args for sum_arg_146
        return(Y_coef_cp[i284]*diff(X_coef_cp[n-i284],True,2)+diff(Y_coef_cp[i284],True,1)*diff(X_coef_cp[n-i284],True,1))

    def sum_arg_145(i284):
        # Child args for sum_arg_145
        return(Y_coef_cp[i284]*diff(X_coef_cp[n-i284],True,1,False,1)+diff(Y_coef_cp[i284],False,1)*diff(X_coef_cp[n-i284],True,1))

    def sum_arg_144(i284):
        # Child args for sum_arg_144
        return(Y_coef_cp[i284]*diff(X_coef_cp[n-i284],True,1))

    def sum_arg_143(i256):
        # Child args for sum_arg_143
        return(i256*diff(X_coef_cp[i256],True,1)*diff(Y_coef_cp[n-i256],False,1)+i256*X_coef_cp[i256]*diff(Y_coef_cp[n-i256],True,1,False,1)+i256*diff(X_coef_cp[i256],False,1)*diff(Y_coef_cp[n-i256],True,1)+i256*diff(X_coef_cp[i256],True,1,False,1)*Y_coef_cp[n-i256])

    def sum_arg_142(i256):
        # Child args for sum_arg_142
        return(i256*X_coef_cp[i256]*diff(Y_coef_cp[n-i256],True,2)+2*i256*diff(X_coef_cp[i256],True,1)*diff(Y_coef_cp[n-i256],True,1)+i256*diff(X_coef_cp[i256],True,2)*Y_coef_cp[n-i256])

    def sum_arg_141(i256):
        # Child args for sum_arg_141
        return(i256*X_coef_cp[i256]*diff(Y_coef_cp[n-i256],True,1)+i256*diff(X_coef_cp[i256],True,1)*Y_coef_cp[n-i256])

    def sum_arg_140(i254):
        # Child args for sum_arg_140
        return((diff(X_coef_cp[i254],True,1)*n-i254*diff(X_coef_cp[i254],True,1))*diff(Y_coef_cp[n-i254],False,1)+(X_coef_cp[i254]*n-i254*X_coef_cp[i254])*diff(Y_coef_cp[n-i254],True,1,False,1)+(diff(X_coef_cp[i254],False,1)*n-i254*diff(X_coef_cp[i254],False,1))*diff(Y_coef_cp[n-i254],True,1)+(diff(X_coef_cp[i254],True,1,False,1)*n-i254*diff(X_coef_cp[i254],True,1,False,1))*Y_coef_cp[n-i254])

    def sum_arg_139(i254):
        # Child args for sum_arg_139
        return((X_coef_cp[i254]*n-i254*X_coef_cp[i254])*diff(Y_coef_cp[n-i254],True,2)+(2*diff(X_coef_cp[i254],True,1)*n-2*i254*diff(X_coef_cp[i254],True,1))*diff(Y_coef_cp[n-i254],True,1)+(diff(X_coef_cp[i254],True,2)*n-i254*diff(X_coef_cp[i254],True,2))*Y_coef_cp[n-i254])

    def sum_arg_138(i254):
        # Child args for sum_arg_138
        return((X_coef_cp[i254]*n-i254*X_coef_cp[i254])*diff(Y_coef_cp[n-i254],True,1)+(diff(X_coef_cp[i254],True,1)*n-i254*diff(X_coef_cp[i254],True,1))*Y_coef_cp[n-i254])

    def sum_arg_137(i348):
        # Child args for sum_arg_137
        def sum_arg_136(i228):
            # Child args for sum_arg_136
            def sum_arg_135(i226):
                # Child args for sum_arg_135
                return(B_denom_coef_c[i226]*B_denom_coef_c[n-i348-i228-i226])

            return(diff(p_perp_coef_cp[i228],False,1)*py_sum(sum_arg_135,0,n-i348-i228))

        return(B_theta_coef_cp[i348]*py_sum(sum_arg_136,0,n-i348))

    def sum_arg_134(i345):
        # Child args for sum_arg_134
        def sum_arg_133(i346):
            # Child args for sum_arg_133
            def sum_arg_132(i344):
                # Child args for sum_arg_132
                return(Delta_coef_cp[i344]*diff(B_theta_coef_cp[(-n)-i346+2*i345-i344],True,1)*is_seq(n-i345,(-i346)+i345-i344))

            return(B_denom_coef_c[i346]*py_sum(sum_arg_132,0,i345-i346))

        return(is_seq(0,n-i345)*iota_coef[n-i345]*is_integer(n-i345)*py_sum(sum_arg_133,0,i345))

    def sum_arg_131(i342):
        # Child args for sum_arg_131
        def sum_arg_130(i340):
            # Child args for sum_arg_130
            return(Delta_coef_cp[i340]*diff(B_theta_coef_cp[n-i342-i340],False,1))

        return(B_denom_coef_c[i342]*py_sum(sum_arg_130,0,n-i342))

    def sum_arg_129(i334):
        # Child args for sum_arg_129
        def sum_arg_127(i296):
            # Child args for sum_arg_127
            return(B_psi_coef_cp[i296]*diff(Delta_coef_cp[n-i334-i296-2],False,1))

        def sum_arg_128(i296):
            # Child args for sum_arg_128
            return(diff(B_psi_coef_cp[i296],True,1)*diff(Delta_coef_cp[n-i334-i296-2],False,1)+B_psi_coef_cp[i296]*diff(Delta_coef_cp[n-i334-i296-2],True,1,False,1))

        return(B_denom_coef_c[i334]*py_sum(sum_arg_128,0,n-i334-2)+diff(B_denom_coef_c[i334],True,1)*py_sum(sum_arg_127,0,n-i334-2))

    def sum_arg_126(i330):
        # Child args for sum_arg_126
        def sum_arg_124(i328):
            # Child args for sum_arg_124
            return(Delta_coef_cp[i328]*diff(B_psi_coef_cp[n-i330-i328-2],False,1))

        def sum_arg_125(i328):
            # Child args for sum_arg_125
            return(diff(Delta_coef_cp[i328],True,1)*diff(B_psi_coef_cp[n-i330-i328-2],False,1)+Delta_coef_cp[i328]*diff(B_psi_coef_cp[n-i330-i328-2],True,1,False,1))

        return(B_denom_coef_c[i330]*py_sum(sum_arg_125,0,n-i330-2)+diff(B_denom_coef_c[i330],True,1)*py_sum(sum_arg_124,0,n-i330-2))

    def sum_arg_123(i337):
        # Child args for sum_arg_123
        def sum_arg_122(i338):
            # Child args for sum_arg_122
            return(B_denom_coef_c[i338]*diff(B_theta_coef_cp[(-n)-i338+2*i337],True,1)*is_seq(n-i337,i337-i338))

        return(is_seq(0,n-i337)*iota_coef[n-i337]*is_integer(n-i337)*py_sum(sum_arg_122,0,i337))

    def sum_arg_121(i336):
        # Child args for sum_arg_121
        return(B_denom_coef_c[i336]*diff(B_theta_coef_cp[n-i336],False,1))

    def sum_arg_120(i325):
        # Child args for sum_arg_120
        def sum_arg_119(i326):
            # Child args for sum_arg_119
            def sum_arg_117(i298):
                # Child args for sum_arg_117
                return(B_psi_coef_cp[i298]*diff(Delta_coef_cp[(-n)-i326+2*i325-i298+2],True,1))

            def sum_arg_118(i298):
                # Child args for sum_arg_118
                return(B_psi_coef_cp[i298]*diff(Delta_coef_cp[(-n)-i326+2*i325-i298+2],True,2)+diff(B_psi_coef_cp[i298],True,1)*diff(Delta_coef_cp[(-n)-i326+2*i325-i298+2],True,1))

            return(is_seq(n-i325-2,i325-i326)*(B_denom_coef_c[i326]*py_sum(sum_arg_118,0,(-n)-i326+2*i325+2)+diff(B_denom_coef_c[i326],True,1)*py_sum(sum_arg_117,0,(-n)-i326+2*i325+2)))

        return(is_seq(0,n-i325-2)*iota_coef[n-i325-2]*is_integer(n-i325-2)*py_sum(sum_arg_119,0,i325))

    def sum_arg_116(i321):
        # Child args for sum_arg_116
        def sum_arg_115(i322):
            # Child args for sum_arg_115
            def sum_arg_113(i318):
                # Child args for sum_arg_113
                return(Delta_coef_cp[i318]*diff(B_psi_coef_cp[(-n)-i322+2*i321-i318+2],True,1))

            def sum_arg_114(i318):
                # Child args for sum_arg_114
                return(Delta_coef_cp[i318]*diff(B_psi_coef_cp[(-n)-i322+2*i321-i318+2],True,2)+diff(Delta_coef_cp[i318],True,1)*diff(B_psi_coef_cp[(-n)-i322+2*i321-i318+2],True,1))

            return(is_seq(n-i321-2,i321-i322)*(B_denom_coef_c[i322]*py_sum(sum_arg_114,0,(-n)-i322+2*i321+2)+diff(B_denom_coef_c[i322],True,1)*py_sum(sum_arg_113,0,(-n)-i322+2*i321+2)))

        return(is_seq(0,n-i321-2)*iota_coef[n-i321-2]*is_integer(n-i321-2)*py_sum(sum_arg_115,0,i321))

    def sum_arg_112(i311):
        # Child args for sum_arg_112
        def sum_arg_111(i312):
            # Child args for sum_arg_111
            return((diff(B_denom_coef_c[i312],True,1)*Delta_coef_cp[(-n)-i312+2*i311]+B_denom_coef_c[i312]*diff(Delta_coef_cp[(-n)-i312+2*i311],True,1))*is_seq(n-i311,i311-i312))

        return((is_seq(0,n-i311)*n-is_seq(0,n-i311)*i311)*B_alpha_coef[n-i311]*is_integer(n-i311)*py_sum(sum_arg_111,0,i311))

    def sum_arg_110(i305):
        # Child args for sum_arg_110
        def sum_arg_109(i306):
            # Child args for sum_arg_109
            return((B_denom_coef_c[i306]*diff(B_psi_coef_cp[(-n)-i306+2*i305+2],True,2)+diff(B_denom_coef_c[i306],True,1)*diff(B_psi_coef_cp[(-n)-i306+2*i305+2],True,1))*is_seq(n-i305-2,i305-i306))

        return(is_seq(0,n-i305-2)*iota_coef[n-i305-2]*is_integer(n-i305-2)*py_sum(sum_arg_109,0,i305))

    def sum_arg_108(i304):
        # Child args for sum_arg_108
        return(diff(B_denom_coef_c[i304],True,1)*diff(B_psi_coef_cp[n-i304-2],False,1)+B_denom_coef_c[i304]*diff(B_psi_coef_cp[n-i304-2],True,1,False,1))

    def sum_arg_107(i299):
        # Child args for sum_arg_107
        def sum_arg_106(i300):
            # Child args for sum_arg_106
            return((i300*diff(B_denom_coef_c[i300],True,1)*Delta_coef_cp[(-n)-i300+2*i299]+i300*B_denom_coef_c[i300]*diff(Delta_coef_cp[(-n)-i300+2*i299],True,1))*is_seq(n-i299,i299-i300))

        return(is_seq(0,n-i299)*B_alpha_coef[n-i299]*is_integer(n-i299)*py_sum(sum_arg_106,0,i299))

    def sum_arg_105(i291):
        # Child args for sum_arg_105
        def sum_arg_104(i292):
            # Child args for sum_arg_104
            def sum_arg_102(i246):
                # Child args for sum_arg_102
                return(B_denom_coef_c[i246]*B_denom_coef_c[(-n)-i292+2*i291-i246])

            def sum_arg_103(i246):
                # Child args for sum_arg_103
                return(diff(B_denom_coef_c[i246],True,1)*B_denom_coef_c[(-n)-i292+2*i291-i246]+B_denom_coef_c[i246]*diff(B_denom_coef_c[(-n)-i292+2*i291-i246],True,1))

            return(is_seq(n-i291,i291-i292)*(i292*p_perp_coef_cp[i292]*py_sum(sum_arg_103,0,(-n)-i292+2*i291)+i292*diff(p_perp_coef_cp[i292],True,1)*py_sum(sum_arg_102,0,(-n)-i292+2*i291)))

        return(is_seq(0,n-i291)*B_alpha_coef[n-i291]*is_integer(n-i291)*py_sum(sum_arg_104,0,i291))

    def sum_arg_101(i290):
        # Child args for sum_arg_101
        return(X_coef_cp[i290]*diff(Z_coef_cp[n-i290],True,2)+diff(X_coef_cp[i290],True,1)*diff(Z_coef_cp[n-i290],True,1))

    def sum_arg_100(i290):
        # Child args for sum_arg_100
        return(X_coef_cp[i290]*diff(Z_coef_cp[n-i290],True,1,False,1)+diff(X_coef_cp[i290],False,1)*diff(Z_coef_cp[n-i290],True,1))

    def sum_arg_99(i290):
        # Child args for sum_arg_99
        return(X_coef_cp[i290]*diff(Z_coef_cp[n-i290],True,1))

    def sum_arg_98(i286):
        # Child args for sum_arg_98
        return(Z_coef_cp[i286]*diff(X_coef_cp[n-i286],True,2)+diff(Z_coef_cp[i286],True,1)*diff(X_coef_cp[n-i286],True,1))

    def sum_arg_97(i286):
        # Child args for sum_arg_97
        return(Z_coef_cp[i286]*diff(X_coef_cp[n-i286],True,1,False,1)+diff(Z_coef_cp[i286],False,1)*diff(X_coef_cp[n-i286],True,1))

    def sum_arg_96(i286):
        # Child args for sum_arg_96
        return(Z_coef_cp[i286]*diff(X_coef_cp[n-i286],True,1))

    def sum_arg_95(i281):
        # Child args for sum_arg_95
        def sum_arg_94(i282):
            # Child args for sum_arg_94
            return((diff(Z_coef_cp[i282],True,1)*diff(Z_coef_cp[(-n)-i282+2*i281],True,2)+diff(Z_coef_cp[i282],True,2)*diff(Z_coef_cp[(-n)-i282+2*i281],True,1))*is_seq(n-i281,i281-i282))

        return(is_seq(0,n-i281)*iota_coef[n-i281]*is_integer(n-i281)*py_sum(sum_arg_94,0,i281))

    def sum_arg_93(i281):
        # Child args for sum_arg_93
        def sum_arg_92(i282):
            # Child args for sum_arg_92
            return((diff(Z_coef_cp[i282],True,1)*diff(Z_coef_cp[(-n)-i282+2*i281],True,1,False,1)+diff(Z_coef_cp[i282],True,1,False,1)*diff(Z_coef_cp[(-n)-i282+2*i281],True,1))*is_seq(n-i281,i281-i282))

        return(is_seq(0,n-i281)*iota_coef[n-i281]*is_integer(n-i281)*py_sum(sum_arg_92,0,i281))

    def sum_arg_91(i281):
        # Child args for sum_arg_91
        def sum_arg_90(i282):
            # Child args for sum_arg_90
            return(diff(Z_coef_cp[i282],True,1)*diff(Z_coef_cp[(-n)-i282+2*i281],True,1)*is_seq(n-i281,i281-i282))

        return(is_seq(0,n-i281)*iota_coef[n-i281]*is_integer(n-i281)*py_sum(sum_arg_90,0,i281))

    def sum_arg_89(i280):
        # Child args for sum_arg_89
        return(diff(Z_coef_cp[i280],True,1)*diff(Z_coef_cp[n-i280],False,2)+diff(Z_coef_cp[i280],True,1,False,1)*diff(Z_coef_cp[n-i280],False,1))

    def sum_arg_88(i280):
        # Child args for sum_arg_88
        return(diff(Z_coef_cp[i280],True,2)*diff(Z_coef_cp[n-i280],False,1)+diff(Z_coef_cp[i280],True,1)*diff(Z_coef_cp[n-i280],True,1,False,1))

    def sum_arg_87(i280):
        # Child args for sum_arg_87
        return(diff(Z_coef_cp[i280],True,1)*diff(Z_coef_cp[n-i280],False,1))

    def sum_arg_86(i277):
        # Child args for sum_arg_86
        def sum_arg_85(i278):
            # Child args for sum_arg_85
            return((diff(Y_coef_cp[i278],True,1)*diff(Y_coef_cp[(-n)-i278+2*i277],True,2)+diff(Y_coef_cp[i278],True,2)*diff(Y_coef_cp[(-n)-i278+2*i277],True,1))*is_seq(n-i277,i277-i278))

        return(is_seq(0,n-i277)*iota_coef[n-i277]*is_integer(n-i277)*py_sum(sum_arg_85,0,i277))

    def sum_arg_84(i277):
        # Child args for sum_arg_84
        def sum_arg_83(i278):
            # Child args for sum_arg_83
            return((diff(Y_coef_cp[i278],True,1)*diff(Y_coef_cp[(-n)-i278+2*i277],True,1,False,1)+diff(Y_coef_cp[i278],True,1,False,1)*diff(Y_coef_cp[(-n)-i278+2*i277],True,1))*is_seq(n-i277,i277-i278))

        return(is_seq(0,n-i277)*iota_coef[n-i277]*is_integer(n-i277)*py_sum(sum_arg_83,0,i277))

    def sum_arg_82(i277):
        # Child args for sum_arg_82
        def sum_arg_81(i278):
            # Child args for sum_arg_81
            return(diff(Y_coef_cp[i278],True,1)*diff(Y_coef_cp[(-n)-i278+2*i277],True,1)*is_seq(n-i277,i277-i278))

        return(is_seq(0,n-i277)*iota_coef[n-i277]*is_integer(n-i277)*py_sum(sum_arg_81,0,i277))

    def sum_arg_80(i276):
        # Child args for sum_arg_80
        return(diff(Y_coef_cp[i276],True,1)*diff(Y_coef_cp[n-i276],False,2)+diff(Y_coef_cp[i276],True,1,False,1)*diff(Y_coef_cp[n-i276],False,1))

    def sum_arg_79(i276):
        # Child args for sum_arg_79
        return(diff(Y_coef_cp[i276],True,2)*diff(Y_coef_cp[n-i276],False,1)+diff(Y_coef_cp[i276],True,1)*diff(Y_coef_cp[n-i276],True,1,False,1))

    def sum_arg_78(i276):
        # Child args for sum_arg_78
        return(diff(Y_coef_cp[i276],True,1)*diff(Y_coef_cp[n-i276],False,1))

    def sum_arg_77(i273):
        # Child args for sum_arg_77
        def sum_arg_76(i274):
            # Child args for sum_arg_76
            return((diff(X_coef_cp[i274],True,1)*diff(X_coef_cp[(-n)-i274+2*i273],True,2)+diff(X_coef_cp[i274],True,2)*diff(X_coef_cp[(-n)-i274+2*i273],True,1))*is_seq(n-i273,i273-i274))

        return(is_seq(0,n-i273)*iota_coef[n-i273]*is_integer(n-i273)*py_sum(sum_arg_76,0,i273))

    def sum_arg_75(i273):
        # Child args for sum_arg_75
        def sum_arg_74(i274):
            # Child args for sum_arg_74
            return((diff(X_coef_cp[i274],True,1)*diff(X_coef_cp[(-n)-i274+2*i273],True,1,False,1)+diff(X_coef_cp[i274],True,1,False,1)*diff(X_coef_cp[(-n)-i274+2*i273],True,1))*is_seq(n-i273,i273-i274))

        return(is_seq(0,n-i273)*iota_coef[n-i273]*is_integer(n-i273)*py_sum(sum_arg_74,0,i273))

    def sum_arg_73(i273):
        # Child args for sum_arg_73
        def sum_arg_72(i274):
            # Child args for sum_arg_72
            return(diff(X_coef_cp[i274],True,1)*diff(X_coef_cp[(-n)-i274+2*i273],True,1)*is_seq(n-i273,i273-i274))

        return(is_seq(0,n-i273)*iota_coef[n-i273]*is_integer(n-i273)*py_sum(sum_arg_72,0,i273))

    def sum_arg_71(i272):
        # Child args for sum_arg_71
        return(diff(X_coef_cp[i272],True,1)*diff(X_coef_cp[n-i272],False,2)+diff(X_coef_cp[i272],True,1,False,1)*diff(X_coef_cp[n-i272],False,1))

    def sum_arg_70(i272):
        # Child args for sum_arg_70
        return(diff(X_coef_cp[i272],True,2)*diff(X_coef_cp[n-i272],False,1)+diff(X_coef_cp[i272],True,1)*diff(X_coef_cp[n-i272],True,1,False,1))

    def sum_arg_69(i272):
        # Child args for sum_arg_69
        return(diff(X_coef_cp[i272],True,1)*diff(X_coef_cp[n-i272],False,1))

    def sum_arg_68(i269):
        # Child args for sum_arg_68
        def sum_arg_67(i270):
            # Child args for sum_arg_67
            return((i270*Z_coef_cp[i270]*diff(Z_coef_cp[(-n)-i270+2*i269],True,3)+2*i270*diff(Z_coef_cp[i270],True,1)*diff(Z_coef_cp[(-n)-i270+2*i269],True,2)+i270*diff(Z_coef_cp[i270],True,2)*diff(Z_coef_cp[(-n)-i270+2*i269],True,1))*is_seq(n-i269,i269-i270))

        return(is_seq(0,n-i269)*iota_coef[n-i269]*is_integer(n-i269)*py_sum(sum_arg_67,0,i269))

    def sum_arg_66(i269):
        # Child args for sum_arg_66
        def sum_arg_65(i270):
            # Child args for sum_arg_65
            return((i270*Z_coef_cp[i270]*diff(Z_coef_cp[(-n)-i270+2*i269],True,2,False,1)+i270*diff(Z_coef_cp[i270],False,1)*diff(Z_coef_cp[(-n)-i270+2*i269],True,2)+i270*diff(Z_coef_cp[i270],True,1)*diff(Z_coef_cp[(-n)-i270+2*i269],True,1,False,1)+i270*diff(Z_coef_cp[i270],True,1,False,1)*diff(Z_coef_cp[(-n)-i270+2*i269],True,1))*is_seq(n-i269,i269-i270))

        return(is_seq(0,n-i269)*iota_coef[n-i269]*is_integer(n-i269)*py_sum(sum_arg_65,0,i269))

    def sum_arg_64(i269):
        # Child args for sum_arg_64
        def sum_arg_63(i270):
            # Child args for sum_arg_63
            return((i270*Z_coef_cp[i270]*diff(Z_coef_cp[(-n)-i270+2*i269],True,2)+i270*diff(Z_coef_cp[i270],True,1)*diff(Z_coef_cp[(-n)-i270+2*i269],True,1))*is_seq(n-i269,i269-i270))

        return(is_seq(0,n-i269)*iota_coef[n-i269]*is_integer(n-i269)*py_sum(sum_arg_63,0,i269))

    def sum_arg_62(i267):
        # Child args for sum_arg_62
        def sum_arg_61(i268):
            # Child args for sum_arg_61
            return((i268*Y_coef_cp[i268]*diff(Y_coef_cp[(-n)-i268+2*i267],True,3)+2*i268*diff(Y_coef_cp[i268],True,1)*diff(Y_coef_cp[(-n)-i268+2*i267],True,2)+i268*diff(Y_coef_cp[i268],True,2)*diff(Y_coef_cp[(-n)-i268+2*i267],True,1))*is_seq(n-i267,i267-i268))

        return(is_seq(0,n-i267)*iota_coef[n-i267]*is_integer(n-i267)*py_sum(sum_arg_61,0,i267))

    def sum_arg_60(i267):
        # Child args for sum_arg_60
        def sum_arg_59(i268):
            # Child args for sum_arg_59
            return((i268*Y_coef_cp[i268]*diff(Y_coef_cp[(-n)-i268+2*i267],True,2,False,1)+i268*diff(Y_coef_cp[i268],False,1)*diff(Y_coef_cp[(-n)-i268+2*i267],True,2)+i268*diff(Y_coef_cp[i268],True,1)*diff(Y_coef_cp[(-n)-i268+2*i267],True,1,False,1)+i268*diff(Y_coef_cp[i268],True,1,False,1)*diff(Y_coef_cp[(-n)-i268+2*i267],True,1))*is_seq(n-i267,i267-i268))

        return(is_seq(0,n-i267)*iota_coef[n-i267]*is_integer(n-i267)*py_sum(sum_arg_59,0,i267))

    def sum_arg_58(i267):
        # Child args for sum_arg_58
        def sum_arg_57(i268):
            # Child args for sum_arg_57
            return((i268*Y_coef_cp[i268]*diff(Y_coef_cp[(-n)-i268+2*i267],True,2)+i268*diff(Y_coef_cp[i268],True,1)*diff(Y_coef_cp[(-n)-i268+2*i267],True,1))*is_seq(n-i267,i267-i268))

        return(is_seq(0,n-i267)*iota_coef[n-i267]*is_integer(n-i267)*py_sum(sum_arg_57,0,i267))

    def sum_arg_56(i265):
        # Child args for sum_arg_56
        def sum_arg_55(i266):
            # Child args for sum_arg_55
            return((i266*X_coef_cp[i266]*diff(X_coef_cp[(-n)-i266+2*i265],True,3)+2*i266*diff(X_coef_cp[i266],True,1)*diff(X_coef_cp[(-n)-i266+2*i265],True,2)+i266*diff(X_coef_cp[i266],True,2)*diff(X_coef_cp[(-n)-i266+2*i265],True,1))*is_seq(n-i265,i265-i266))

        return(is_seq(0,n-i265)*iota_coef[n-i265]*is_integer(n-i265)*py_sum(sum_arg_55,0,i265))

    def sum_arg_54(i265):
        # Child args for sum_arg_54
        def sum_arg_53(i266):
            # Child args for sum_arg_53
            return((i266*X_coef_cp[i266]*diff(X_coef_cp[(-n)-i266+2*i265],True,2,False,1)+i266*diff(X_coef_cp[i266],False,1)*diff(X_coef_cp[(-n)-i266+2*i265],True,2)+i266*diff(X_coef_cp[i266],True,1)*diff(X_coef_cp[(-n)-i266+2*i265],True,1,False,1)+i266*diff(X_coef_cp[i266],True,1,False,1)*diff(X_coef_cp[(-n)-i266+2*i265],True,1))*is_seq(n-i265,i265-i266))

        return(is_seq(0,n-i265)*iota_coef[n-i265]*is_integer(n-i265)*py_sum(sum_arg_53,0,i265))

    def sum_arg_52(i265):
        # Child args for sum_arg_52
        def sum_arg_51(i266):
            # Child args for sum_arg_51
            return((i266*X_coef_cp[i266]*diff(X_coef_cp[(-n)-i266+2*i265],True,2)+i266*diff(X_coef_cp[i266],True,1)*diff(X_coef_cp[(-n)-i266+2*i265],True,1))*is_seq(n-i265,i265-i266))

        return(is_seq(0,n-i265)*iota_coef[n-i265]*is_integer(n-i265)*py_sum(sum_arg_51,0,i265))

    def sum_arg_50(i264):
        # Child args for sum_arg_50
        return(i264*diff(Z_coef_cp[i264],True,1)*diff(Z_coef_cp[n-i264],False,2)+i264*diff(Z_coef_cp[i264],True,1,False,1)*diff(Z_coef_cp[n-i264],False,1)+i264*Z_coef_cp[i264]*diff(Z_coef_cp[n-i264],True,1,False,2)+i264*diff(Z_coef_cp[i264],False,1)*diff(Z_coef_cp[n-i264],True,1,False,1))

    def sum_arg_49(i264):
        # Child args for sum_arg_49
        return(i264*diff(Z_coef_cp[i264],True,2)*diff(Z_coef_cp[n-i264],False,1)+i264*Z_coef_cp[i264]*diff(Z_coef_cp[n-i264],True,2,False,1)+2*i264*diff(Z_coef_cp[i264],True,1)*diff(Z_coef_cp[n-i264],True,1,False,1))

    def sum_arg_48(i264):
        # Child args for sum_arg_48
        return(i264*diff(Z_coef_cp[i264],True,1)*diff(Z_coef_cp[n-i264],False,1)+i264*Z_coef_cp[i264]*diff(Z_coef_cp[n-i264],True,1,False,1))

    def sum_arg_47(i262):
        # Child args for sum_arg_47
        return(i262*diff(Y_coef_cp[i262],True,1)*diff(Y_coef_cp[n-i262],False,2)+i262*diff(Y_coef_cp[i262],True,1,False,1)*diff(Y_coef_cp[n-i262],False,1)+i262*Y_coef_cp[i262]*diff(Y_coef_cp[n-i262],True,1,False,2)+i262*diff(Y_coef_cp[i262],False,1)*diff(Y_coef_cp[n-i262],True,1,False,1))

    def sum_arg_46(i262):
        # Child args for sum_arg_46
        return(i262*diff(Y_coef_cp[i262],True,2)*diff(Y_coef_cp[n-i262],False,1)+i262*Y_coef_cp[i262]*diff(Y_coef_cp[n-i262],True,2,False,1)+2*i262*diff(Y_coef_cp[i262],True,1)*diff(Y_coef_cp[n-i262],True,1,False,1))

    def sum_arg_45(i262):
        # Child args for sum_arg_45
        return(i262*diff(Y_coef_cp[i262],True,1)*diff(Y_coef_cp[n-i262],False,1)+i262*Y_coef_cp[i262]*diff(Y_coef_cp[n-i262],True,1,False,1))

    def sum_arg_44(i260):
        # Child args for sum_arg_44
        return(i260*diff(X_coef_cp[i260],True,1)*diff(X_coef_cp[n-i260],False,2)+i260*diff(X_coef_cp[i260],True,1,False,1)*diff(X_coef_cp[n-i260],False,1)+i260*X_coef_cp[i260]*diff(X_coef_cp[n-i260],True,1,False,2)+i260*diff(X_coef_cp[i260],False,1)*diff(X_coef_cp[n-i260],True,1,False,1))

    def sum_arg_43(i260):
        # Child args for sum_arg_43
        return(i260*diff(X_coef_cp[i260],True,2)*diff(X_coef_cp[n-i260],False,1)+i260*X_coef_cp[i260]*diff(X_coef_cp[n-i260],True,2,False,1)+2*i260*diff(X_coef_cp[i260],True,1)*diff(X_coef_cp[n-i260],True,1,False,1))

    def sum_arg_42(i260):
        # Child args for sum_arg_42
        return(i260*diff(X_coef_cp[i260],True,1)*diff(X_coef_cp[n-i260],False,1)+i260*X_coef_cp[i260]*diff(X_coef_cp[n-i260],True,1,False,1))

    def sum_arg_41(i258):
        # Child args for sum_arg_41
        return(i258*diff(X_coef_cp[i258],True,1)*diff(Z_coef_cp[n-i258],False,1)+i258*X_coef_cp[i258]*diff(Z_coef_cp[n-i258],True,1,False,1)+i258*diff(X_coef_cp[i258],False,1)*diff(Z_coef_cp[n-i258],True,1)+i258*diff(X_coef_cp[i258],True,1,False,1)*Z_coef_cp[n-i258])

    def sum_arg_40(i258):
        # Child args for sum_arg_40
        return(i258*X_coef_cp[i258]*diff(Z_coef_cp[n-i258],True,2)+2*i258*diff(X_coef_cp[i258],True,1)*diff(Z_coef_cp[n-i258],True,1)+i258*diff(X_coef_cp[i258],True,2)*Z_coef_cp[n-i258])

    def sum_arg_39(i258):
        # Child args for sum_arg_39
        return(i258*X_coef_cp[i258]*diff(Z_coef_cp[n-i258],True,1)+i258*diff(X_coef_cp[i258],True,1)*Z_coef_cp[n-i258])

    def sum_arg_38(i252):
        # Child args for sum_arg_38
        return((diff(X_coef_cp[i252],True,1)*n-i252*diff(X_coef_cp[i252],True,1))*diff(Z_coef_cp[n-i252],False,1)+(X_coef_cp[i252]*n-i252*X_coef_cp[i252])*diff(Z_coef_cp[n-i252],True,1,False,1)+(diff(X_coef_cp[i252],False,1)*n-i252*diff(X_coef_cp[i252],False,1))*diff(Z_coef_cp[n-i252],True,1)+(diff(X_coef_cp[i252],True,1,False,1)*n-i252*diff(X_coef_cp[i252],True,1,False,1))*Z_coef_cp[n-i252])

    def sum_arg_37(i252):
        # Child args for sum_arg_37
        return((X_coef_cp[i252]*n-i252*X_coef_cp[i252])*diff(Z_coef_cp[n-i252],True,2)+(2*diff(X_coef_cp[i252],True,1)*n-2*i252*diff(X_coef_cp[i252],True,1))*diff(Z_coef_cp[n-i252],True,1)+(diff(X_coef_cp[i252],True,2)*n-i252*diff(X_coef_cp[i252],True,2))*Z_coef_cp[n-i252])

    def sum_arg_36(i252):
        # Child args for sum_arg_36
        return((X_coef_cp[i252]*n-i252*X_coef_cp[i252])*diff(Z_coef_cp[n-i252],True,1)+(diff(X_coef_cp[i252],True,1)*n-i252*diff(X_coef_cp[i252],True,1))*Z_coef_cp[n-i252])

    def sum_arg_35(i209):
        # Child args for sum_arg_35
        def sum_arg_34(i210):
            # Child args for sum_arg_34
            return((diff(B_theta_coef_cp[i210],True,1)*B_denom_coef_c[(-n)-i210+2*i209]+B_theta_coef_cp[i210]*diff(B_denom_coef_c[(-n)-i210+2*i209],True,1))*is_seq(n-i209,i209-i210))

        return(is_seq(0,n-i209)*B_alpha_coef[n-i209]*is_integer(n-i209)*py_sum(sum_arg_34,0,i209))

    def sum_arg_33(i209):
        # Child args for sum_arg_33
        def sum_arg_32(i210):
            # Child args for sum_arg_32
            return(diff(B_theta_coef_cp[i210],False,1)*B_denom_coef_c[(-n)-i210+2*i209]*is_seq(n-i209,i209-i210))

        return(is_seq(0,n-i209)*B_alpha_coef[n-i209]*is_integer(n-i209)*py_sum(sum_arg_32,0,i209))

    def sum_arg_31(i209):
        # Child args for sum_arg_31
        def sum_arg_30(i210):
            # Child args for sum_arg_30
            return(B_theta_coef_cp[i210]*B_denom_coef_c[(-n)-i210+2*i209]*is_seq(n-i209,i209-i210))

        return(is_seq(0,n-i209)*B_alpha_coef[n-i209]*is_integer(n-i209)*py_sum(sum_arg_30,0,i209))

    def sum_arg_29(i201):
        # Child args for sum_arg_29
        def sum_arg_28(i202):
            # Child args for sum_arg_28
            return((diff(B_psi_coef_cp[i202],True,2)*B_denom_coef_c[(-n)-i202+2*i201+2]+B_psi_coef_cp[i202]*diff(B_denom_coef_c[(-n)-i202+2*i201+2],True,2)+2*diff(B_psi_coef_cp[i202],True,1)*diff(B_denom_coef_c[(-n)-i202+2*i201+2],True,1))*is_seq(n-i201-2,i201-i202))

        return(is_seq(0,n-i201-2)*B_alpha_coef[n-i201-2]*is_integer(n-i201-2)*py_sum(sum_arg_28,0,i201))

    def sum_arg_27(i201):
        # Child args for sum_arg_27
        def sum_arg_26(i202):
            # Child args for sum_arg_26
            return((diff(B_psi_coef_cp[i202],True,1,False,1)*B_denom_coef_c[(-n)-i202+2*i201+2]+diff(B_psi_coef_cp[i202],False,1)*diff(B_denom_coef_c[(-n)-i202+2*i201+2],True,1))*is_seq(n-i201-2,i201-i202))

        return(is_seq(0,n-i201-2)*B_alpha_coef[n-i201-2]*is_integer(n-i201-2)*py_sum(sum_arg_26,0,i201))

    def sum_arg_25(i201):
        # Child args for sum_arg_25
        def sum_arg_24(i202):
            # Child args for sum_arg_24
            return((diff(B_psi_coef_cp[i202],True,1)*B_denom_coef_c[(-n)-i202+2*i201+2]+B_psi_coef_cp[i202]*diff(B_denom_coef_c[(-n)-i202+2*i201+2],True,1))*is_seq(n-i201-2,i201-i202))

        return(is_seq(0,n-i201-2)*B_alpha_coef[n-i201-2]*is_integer(n-i201-2)*py_sum(sum_arg_24,0,i201))

    def sum_arg_23(i355):
        # Child args for sum_arg_23
        def sum_arg_22(i356):
            # Child args for sum_arg_22
            def sum_arg_21(i804):
                # Child args for sum_arg_21
                return(diff(B_denom_coef_c[(-i804)-i356+i355],True,1)*Delta_coef_cp[i804])

            return(is_seq(0,(-n)+i356+i355)*B_theta_coef_cp[(-n)+i356+i355]*is_integer((-n)+i356+i355)*is_seq((-n)+i356+i355,i356)*py_sum(sum_arg_21,0,i355-i356))

        return(iota_coef[n-i355]*py_sum(sum_arg_22,0,i355))

    def sum_arg_20(i353):
        # Child args for sum_arg_20
        def sum_arg_19(i238):
            # Child args for sum_arg_19
            return(Delta_coef_cp[i238]*diff(B_denom_coef_c[(-n)+2*i353-i238],True,1))

        return(is_seq(0,n-i353)*B_alpha_coef[n-i353]*is_integer(n-i353)*is_seq(n-i353,i353)*py_sum(sum_arg_19,0,2*i353-n))

    def sum_arg_18(i351):
        # Child args for sum_arg_18
        def sum_arg_17(i352):
            # Child args for sum_arg_17
            def sum_arg_16(i765):
                # Child args for sum_arg_16
                def sum_arg_15(i788):
                    # Child args for sum_arg_15
                    return(B_denom_coef_c[(-i788)-i765-i352+i351]*B_denom_coef_c[i788])

                return(diff(p_perp_coef_cp[i765],True,1)*py_sum(sum_arg_15,0,(-i765)-i352+i351))

            return(is_seq(0,(-n)+i352+i351)*B_theta_coef_cp[(-n)+i352+i351]*is_integer((-n)+i352+i351)*is_seq((-n)+i352+i351,i352)*py_sum(sum_arg_16,0,i351-i352))

        return(iota_coef[n-i351]*py_sum(sum_arg_17,0,i351))

    def sum_arg_14(i349):
        # Child args for sum_arg_14
        def sum_arg_13(i234):
            # Child args for sum_arg_13
            def sum_arg_12(i230):
                # Child args for sum_arg_12
                return(B_denom_coef_c[i230]*B_denom_coef_c[(-n)+2*i349-i234-i230])

            return(diff(p_perp_coef_cp[i234],True,1)*py_sum(sum_arg_12,0,(-n)+2*i349-i234))

        return(is_seq(0,n-i349)*B_alpha_coef[n-i349]*is_integer(n-i349)*is_seq(n-i349,i349)*py_sum(sum_arg_13,0,2*i349-n))

    def sum_arg_11(i331):
        # Child args for sum_arg_11
        def sum_arg_9(i332):
            # Child args for sum_arg_9
            return((is_seq(0,(-n)-i332+2*i331)*diff(B_denom_coef_c[i332],True,1)*B_theta_coef_cp[(-n)-i332+2*i331]+is_seq(0,(-n)-i332+2*i331)*B_denom_coef_c[i332]*diff(B_theta_coef_cp[(-n)-i332+2*i331],True,1))*is_integer((-n)-i332+2*i331)*is_seq((-n)-i332+2*i331,i331-i332))

        def sum_arg_10(i332):
            # Child args for sum_arg_10
            return((is_seq(0,(-n)-i332+2*i331)*diff(B_denom_coef_c[i332],True,1)*B_theta_coef_cp[(-n)-i332+2*i331]+is_seq(0,(-n)-i332+2*i331)*B_denom_coef_c[i332]*diff(B_theta_coef_cp[(-n)-i332+2*i331],True,1))*is_integer((-n)-i332+2*i331)*is_seq((-n)-i332+2*i331,i331-i332))

        return(iota_coef[n-i331]*(n*py_sum(sum_arg_10,0,i331)-i331*py_sum(sum_arg_9,0,i331)))

    def sum_arg_8(i315):
        # Child args for sum_arg_8
        def sum_arg_4(i316):
            # Child args for sum_arg_4
            def sum_arg_2(i314):
                # Child args for sum_arg_2
                return(is_seq(0,(-n)-i316+2*i315-i314)*Delta_coef_cp[i314]*B_theta_coef_cp[(-n)-i316+2*i315-i314]*is_integer((-n)-i316+2*i315-i314)*is_seq((-n)-i316+2*i315-i314,(-i316)+i315-i314))

            def sum_arg_3(i314):
                # Child args for sum_arg_3
                return((is_seq(0,(-n)-i316+2*i315-i314)*diff(Delta_coef_cp[i314],True,1)*B_theta_coef_cp[(-n)-i316+2*i315-i314]+is_seq(0,(-n)-i316+2*i315-i314)*Delta_coef_cp[i314]*diff(B_theta_coef_cp[(-n)-i316+2*i315-i314],True,1))*is_integer((-n)-i316+2*i315-i314)*is_seq((-n)-i316+2*i315-i314,(-i316)+i315-i314))

            return(B_denom_coef_c[i316]*py_sum(sum_arg_3,0,i315-i316)+diff(B_denom_coef_c[i316],True,1)*py_sum(sum_arg_2,0,i315-i316))

        def sum_arg_7(i316):
            # Child args for sum_arg_7
            def sum_arg_5(i314):
                # Child args for sum_arg_5
                return(is_seq(0,(-n)-i316+2*i315-i314)*Delta_coef_cp[i314]*B_theta_coef_cp[(-n)-i316+2*i315-i314]*is_integer((-n)-i316+2*i315-i314)*is_seq((-n)-i316+2*i315-i314,(-i316)+i315-i314))

            def sum_arg_6(i314):
                # Child args for sum_arg_6
                return((is_seq(0,(-n)-i316+2*i315-i314)*diff(Delta_coef_cp[i314],True,1)*B_theta_coef_cp[(-n)-i316+2*i315-i314]+is_seq(0,(-n)-i316+2*i315-i314)*Delta_coef_cp[i314]*diff(B_theta_coef_cp[(-n)-i316+2*i315-i314],True,1))*is_integer((-n)-i316+2*i315-i314)*is_seq((-n)-i316+2*i315-i314,(-i316)+i315-i314))

            return(B_denom_coef_c[i316]*py_sum(sum_arg_6,0,i315-i316)+diff(B_denom_coef_c[i316],True,1)*py_sum(sum_arg_5,0,i315-i316))

        return(iota_coef[n-i315]*(n*py_sum(sum_arg_7,0,i315)-i315*py_sum(sum_arg_4,0,i315)))

    def sum_arg_1(i307):
        # Child args for sum_arg_1
        return((is_seq(0,n-i307)*diff(B_denom_coef_c[2*i307-n],True,1)*n-is_seq(0,n-i307)*i307*diff(B_denom_coef_c[2*i307-n],True,1))*B_alpha_coef[n-i307]*is_integer(n-i307)*is_seq(n-i307,i307))


    out = -(((2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_153,0,n)+(2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_152,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_151,0,n)+(2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_150,0,n))*diff(tau_p,False,1)+((2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_149,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_148,0,n)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_147,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_146,0,n)+(2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_145,0,n)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*n*is_integer(n)*py_sum_parallel(sum_arg_144,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_143,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_142,0,n)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_141,0,n)+(2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_140,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_139,0,n)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_138,0,n))*tau_p+((2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*diff(kap_p,False,1)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*kap_p)*n*is_integer(n)*py_sum_parallel(sum_arg_99,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*dl_p*kap_p*n*is_integer(n)*py_sum_parallel(sum_arg_98,0,n)+(2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*kap_p*n*is_integer(n)*py_sum_parallel(sum_arg_97,0,n)+((2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*diff(kap_p,False,1)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*kap_p)*n*is_integer(n)*py_sum_parallel(sum_arg_96,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*n*py_sum_parallel(sum_arg_95,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*n*py_sum_parallel(sum_arg_93,ceil(n/2),floor(n))-2*diff(Delta_coef_cp[0],False,1)*n*py_sum_parallel(sum_arg_91,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_89,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_88,0,n)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*n*is_integer(n)*py_sum_parallel(sum_arg_87,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*n*py_sum_parallel(sum_arg_86,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*n*py_sum_parallel(sum_arg_84,ceil(n/2),floor(n))-2*diff(Delta_coef_cp[0],False,1)*n*py_sum_parallel(sum_arg_82,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_80,0,n)+4*B_alpha_coef[0]*py_sum_parallel(sum_arg_8,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_79,0,n)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*n*is_integer(n)*py_sum_parallel(sum_arg_78,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*n*py_sum_parallel(sum_arg_77,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*n*py_sum_parallel(sum_arg_75,ceil(n/2),floor(n))-2*diff(Delta_coef_cp[0],False,1)*n*py_sum_parallel(sum_arg_73,ceil(n/2),floor(n))+(2-2*Delta_coef_cp[0])*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_71,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_70,0,n)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*n*is_integer(n)*py_sum_parallel(sum_arg_69,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*py_sum_parallel(sum_arg_68,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*py_sum_parallel(sum_arg_66,ceil(n/2),floor(n))+2*diff(Delta_coef_cp[0],False,1)*py_sum_parallel(sum_arg_64,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*iota_coef[0]*py_sum_parallel(sum_arg_62,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*py_sum_parallel(sum_arg_60,ceil(n/2),floor(n))+2*diff(Delta_coef_cp[0],False,1)*py_sum_parallel(sum_arg_58,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*iota_coef[0]*py_sum_parallel(sum_arg_56,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*py_sum_parallel(sum_arg_54,ceil(n/2),floor(n))+2*diff(Delta_coef_cp[0],False,1)*py_sum_parallel(sum_arg_52,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_50,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_49,0,n)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*is_integer(n)*py_sum_parallel(sum_arg_48,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_47,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_46,0,n)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*is_integer(n)*py_sum_parallel(sum_arg_45,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_44,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_43,0,n)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*is_integer(n)*py_sum_parallel(sum_arg_42,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_41,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_40,0,n)+((2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*diff(kap_p,False,1)+2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*kap_p)*is_integer(n)*py_sum_parallel(sum_arg_39,0,n)+(2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_38,0,n)+(2-2*Delta_coef_cp[0])*iota_coef[0]*is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_37,0,n)+((2-2*Delta_coef_cp[0])*is_seq(0,n)*dl_p*diff(kap_p,False,1)-2*is_seq(0,n)*diff(Delta_coef_cp[0],False,1)*dl_p*kap_p)*is_integer(n)*py_sum_parallel(sum_arg_36,0,n)+(2*Delta_coef_cp[0]-2)*iota_coef[0]*n*py_sum_parallel(sum_arg_35,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*n*py_sum_parallel(sum_arg_33,ceil(n/2),floor(n))+2*diff(Delta_coef_cp[0],False,1)*n*py_sum_parallel(sum_arg_31,ceil(n/2),floor(n))+(4-4*Delta_coef_cp[0])*iota_coef[0]*py_sum_parallel(sum_arg_29,ceil(n/2)-1,floor(n)-2)+(4-4*Delta_coef_cp[0])*py_sum_parallel(sum_arg_27,ceil(n/2)-1,floor(n)-2)-4*diff(Delta_coef_cp[0],False,1)*py_sum_parallel(sum_arg_25,ceil(n/2)-1,floor(n)-2)+n*((-B_alpha_coef[0]*py_sum_parallel(sum_arg_23,ceil(n/2),floor(n)))+B_alpha_coef[0]*py_sum_parallel(sum_arg_20,ceil(n/2),floor(n))+2*B_alpha_coef[0]*py_sum_parallel(sum_arg_18,ceil(n/2),floor(n))-2*B_alpha_coef[0]*py_sum_parallel(sum_arg_14,ceil(n/2),floor(n)))+2*B_alpha_coef[0]*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_137,0,n)-2*B_alpha_coef[0]*n*py_sum_parallel(sum_arg_134,ceil(n/2),floor(n))-2*B_alpha_coef[0]*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_131,0,n)+4*B_alpha_coef[0]*is_seq(0,n-2)*is_integer(n-2)*py_sum_parallel(sum_arg_129,0,n-2)+4*B_alpha_coef[0]*is_seq(0,n-2)*is_integer(n-2)*py_sum_parallel(sum_arg_126,0,n-2)+2*B_alpha_coef[0]*n*py_sum_parallel(sum_arg_123,ceil(n/2),floor(n))+2*B_alpha_coef[0]*is_seq(0,n)*n*is_integer(n)*py_sum_parallel(sum_arg_121,0,n)+4*B_alpha_coef[0]*py_sum_parallel(sum_arg_120,ceil(n/2)-1,floor(n)-2)+4*B_alpha_coef[0]*py_sum_parallel(sum_arg_116,ceil(n/2)-1,floor(n)-2)-4*B_alpha_coef[0]*py_sum_parallel(sum_arg_112,ceil(n/2),floor(n))-4*B_alpha_coef[0]*py_sum_parallel(sum_arg_110,ceil(n/2)-1,floor(n)-2)-4*B_alpha_coef[0]*py_sum_parallel(sum_arg_11,ceil(n/2),floor(n))-4*B_alpha_coef[0]*is_seq(0,n-2)*is_integer(n-2)*py_sum_parallel(sum_arg_108,0,n-2)-B_alpha_coef[0]*py_sum_parallel(sum_arg_107,ceil(n/2),floor(n))+2*B_alpha_coef[0]*py_sum_parallel(sum_arg_105,ceil(n/2),floor(n))+(2*Delta_coef_cp[0]-2)*iota_coef[0]*is_seq(0,n)*dl_p*kap_p*n*is_integer(n)*py_sum_parallel(sum_arg_101,0,n)+(2*Delta_coef_cp[0]-2)*is_seq(0,n)*dl_p*kap_p*n*is_integer(n)*py_sum_parallel(sum_arg_100,0,n)+4*B_alpha_coef[0]*py_sum_parallel(sum_arg_1,ceil(n/2),floor(n)))/(2*B_alpha_coef[0]*n)
    return(out)



# Wrapped/completed recursion relations based on translated expressions
# in parsed/. Necessary masking and/or n-substitution are included. All iterate_*
# methods returns ChiPhiFunc's.
import jax.numpy as jnp
import numpy as np # used in save_plain and get_helicity
# from jax import jit, vmap, tree_util
from jax import tree_util
# from functools import partial # for JAX jit with static params
# from matplotlib import pyplot as plt

# ChiPhiFunc and ChiPhiEpsFunc
from .chiphifunc import *
from .chiphiepsfunc import *
from .math_utilities import *
import aqsc.looped_solver as looped_solver

# parsed relations
import aqsc.parsed as parsed
import aqsc.MHD_parsed as MHD_parsed
import aqsc.looped_coefs as looped_coefs

# Plotting
import matplotlib.pyplot as plt
from matplotlib import cm, colors

''' I. Magnetic equations '''
# The magnetic equations alone can serve as a set of recursion relations,
# solving for $X, Y, Z, B_{\psi}, \iota$ from $B_{\theta}, B_{alpha}$ and $B$.
''' I.1 Recursion relations for individual variables '''
# All iterate_<variable name> functions evaluate said variable at order n_eval.
# See overleaf (will be offline later) document which variables are needed and
# which orders are needed.
# not nfp-dependent
def iterate_Xn_cp(n_eval,
    X_coef_cp,
    Y_coef_cp,
    Z_coef_cp,
    B_denom_coef_c,
    B_alpha_coef,
    kap_p, dl_p, tau_p,
    iota_coef):
    return(parsed.eval_Xn_cp(
        n=n_eval,
        X_coef_cp=X_coef_cp.mask(n_eval-1).zero_append(),
        Y_coef_cp=Y_coef_cp.mask(n_eval-1),
        Z_coef_cp=Z_coef_cp.mask(n_eval),
        B_denom_coef_c=B_denom_coef_c,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p, dl_p=dl_p, tau_p=tau_p,
        iota_coef=iota_coef).cap_m(n_eval))

# O_matrices, O_einv, vector_free_coef only uses B_alpha_coef and X_coef_cp
def iterate_Yn_cp_operators(n_unknown, X_coef_cp, B_alpha_coef, Y1c_mode=False): # nfp-dependent only in output
    '''
    Input: -----
    '''
    # Getting coeffs
    # Both uses B_alpha0 and X1 only
    chiphifunc_A = parsed.coef_a(n_unknown-1, B_alpha_coef, X_coef_cp)
    chiphifunc_B = parsed.coef_b(B_alpha_coef, X_coef_cp)

    # Calculating the inverted matrices
    O_matrices, O_einv, vector_free_coef = get_O_O_einv_from_A_B(
        chiphifunc_A=chiphifunc_A,
        chiphifunc_B=chiphifunc_B,
        rank_rhs=n_unknown,
        Y1c_mode=Y1c_mode)
    return(O_matrices, O_einv, vector_free_coef)

# O_matrices, O_einv, vector_free_coef only uses B_alpha_coef and X_coef_cp
# nfp-dependent!!
def iterate_Yn_cp_RHS(n_unknown,
    X_coef_cp,
    Y_coef_cp,
    Z_coef_cp,
    B_psi_coef_cp,
    B_theta_coef_cp,
    B_alpha_coef,
    B_denom_coef_c,
    kap_p, dl_p, tau_p,
    iota_coef): # nfp-dependent!!
    # This result is known due to indexing.
    if n_unknown == 1:
        return(ChiPhiFunc(jnp.zeros((3,X_coef_cp[1].content.shape[1])), X_coef_cp.nfp))
    # Getting rhs-lhs for the Yn+1 equation
    # for Yn to work, "n" must be subbed with n-1 here
    chiphifunc_rhs = parsed.rhs_minus_lhs(n_unknown-1,
        X_coef_cp,
        Y_coef_cp.mask(n_unknown-1).zero_append(),
        Z_coef_cp,
        B_psi_coef_cp,
        B_theta_coef_cp,
        B_alpha_coef,
        kap_p, dl_p, tau_p,
        iota_coef)

    # Making sure RHS isn't null
    if not isinstance(chiphifunc_rhs, ChiPhiFunc):
        return()

    len_chi = chiphifunc_rhs.content.shape[0]
    # Center-pad chiphifunc_rhs if it's too short
    # ChiPhiFunc_A always have 2 chi components.
    #          [0  ]
    # [RHS] => [RHS]
    #          [0  ]
    if n_unknown+2 != len_chi:
        chiphifunc_rhs = chiphifunc_rhs+\
            ChiPhiFunc(jnp.zeros(
                (
                    n_unknown+2,
                    chiphifunc_rhs.content.shape[1]
                ), dtype=jnp.complex128
            ), chiphifunc_rhs.nfp)

    return(chiphifunc_rhs)

# Evaluates Yn using Yn+1 general formula. The free component is either Yn0 or
# Yn1s.
# Requires:
# X_{n}, Y_{n-1}, Z_{n-1},
# B_{\theta n-1}, B_{\psi  n-3},
# \iota_{(n-3)/2 or (n-4)/2}, B_{\alpha  (n-1)/2 or (n-2)/2}
# nfp-dependent!!
# Cannot be jitted beause of the lambda funcs
def iterate_Yn_cp_magnetic(
    n_unknown,
    X_coef_cp,
    Y_coef_cp,
    Z_coef_cp,
    B_psi_coef_cp,
    B_theta_coef_cp,
    B_alpha_coef,
    B_denom_coef_c,
    kap_p, dl_p, tau_p,
    iota_coef,
    static_max_freq,
    Yn0=None,
    Yn1c_avg=None):

    nfp = X_coef_cp.nfp
    n_eval = n_unknown+1

    _, O_einv, vector_free_coef = \
        iterate_Yn_cp_operators(
            n_unknown,
            X_coef_cp=X_coef_cp,
            B_alpha_coef=B_alpha_coef, 
            Y1c_mode=True
        )

    Yn_rhs = iterate_Yn_cp_RHS(n_unknown=n_unknown,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp.zero_append(),
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef)
    Yn_rhs_content = Yn_rhs.content
    new_Y_n_no_unknown = ChiPhiFunc(jnp.einsum('ijk,jk->ik',O_einv,Yn_rhs_content), Yn_rhs.nfp)
    Y_coef_cp_no_unknown = Y_coef_cp.mask(n_eval-2)
    Y_coef_cp_no_unknown = Y_coef_cp_no_unknown.append(new_Y_n_no_unknown)

    # Calculating D3 to solve for Yn1p
    if n_unknown%2==0:
        if Yn0 is None:
            return(ChiPhiFuncSpecial(-15))
        Yn_free_content = Yn0
    else:
        if Yn1c_avg is None:
            Yn1c_avg = 0
        D3_RHS_no_unknown = -parsed.eval_D3_RHS_m_LHS(
            n = n_eval,
            X_coef_cp = X_coef_cp.mask(n_unknown).zero_append(),
            # Only dep on Y[+-1]
            Y_coef_cp = Y_coef_cp_no_unknown.zero_append(),
            # The m=0 component is actually indep of Z[n+1]
            Z_coef_cp = Z_coef_cp.mask(n_unknown).zero_append(),
            # This equation may contain both B_theta[n,+-1] and B_theta[n+1,0].
            B_theta_coef_cp = B_theta_coef_cp.mask(n_unknown+1).zero_append(),
            B_denom_coef_c = B_denom_coef_c,
            B_alpha_coef = B_alpha_coef,
            iota_coef = iota_coef, #.mask((n_unknown-3)//2).zero_append(), # iota is also masked
            dl_p = dl_p,
            tau_p = tau_p,
        kap_p = kap_p)[0]

        coef_Yn1c_in_D3, coef_dp_Yn1c_in_D3 = looped_coefs.lambda_coef_Yn1c_in_D3(
            vector_free_coef, 
            X_coef_cp, Y_coef_cp, 
            iota_coef, 
            dl_p, tau_p, 
            nfp
        )
        # Solving y'+py=f for Yn1+. This equation has no unique solution,
        # and a initial condition is provided.
        p_eff = (coef_Yn1c_in_D3.content/coef_dp_Yn1c_in_D3.content)[0]
        f_eff = (D3_RHS_no_unknown.content/coef_dp_Yn1c_in_D3.content)[0]
        p_fft = fft_filter(jnp.fft.fft(p_eff), static_max_freq, axis=0)
        f_fft = fft_filter(jnp.fft.fft(f_eff), static_max_freq, axis=0)
        # Creating differential operator and convolution operator
        # as in solve_ODE
        diff_matrix = fft_dphi_op(static_max_freq)
        conv_matrix = fft_conv_op(p_fft)
        tot_matrix = diff_matrix + conv_matrix
        
        tot_matrix_normalization = jnp.max(jnp.abs(tot_matrix))
        tot_matrix = tot_matrix.at[:, 0].set(
            tot_matrix[:, 0]+tot_matrix_normalization # was +1
        )
        f_fft = f_fft+Yn1c_avg*static_max_freq*tot_matrix_normalization
        sln_fft = jnp.linalg.solve(tot_matrix, f_fft)

        
        len_phi = Yn_rhs.content.shape[1]
        Yn_free_content = jnp.fft.ifft(fft_pad(sln_fft, len_phi, axis=0), axis=0)[None, :]
        
        ######## Yn1p
        # Yn_free_content = solve_ODE(
        #     coeff_arr=coef_Yn1p_in_D3.content,
        #     coeff_dp_arr=coef_dp_Yn1p_in_D3.content*nfp,
        #     f_arr=D3_RHS_no_unknown.content,
        #     static_max_freq=static_max_freq
        # ) # Seems underdetermined. Blows up.


        # # Solving y'+py=f for Yn1+. This equation has no unique solution,
        # # and a initial condition is provided.
        # p_eff = (coef_Yn1p_in_D3.content/coef_dp_Yn1p_in_D3.content)[0]
        # f_eff = (D3_RHS_no_unknown.content/coef_dp_Yn1p_in_D3.content)[0]
        # p_fft = fft_filter(jnp.fft.fft(p_eff), static_max_freq, axis=0)
        # f_fft = fft_filter(jnp.fft.fft(f_eff), static_max_freq, axis=0)
        # # Creating differential operator and convolution operator
        # # as in solve_ODE
        # diff_matrix = fft_dphi_op(static_max_freq)
        # conv_matrix = fft_conv_op(p_fft)
        # tot_matrix = diff_matrix + conv_matrix
        
        # tot_matrix_normalization = jnp.max(jnp.abs(tot_matrix))
        # tot_matrix = tot_matrix.at[:, 0].set(
        #     tot_matrix[:, 0]+tot_matrix_normalization # was +1
        # )
        # f_fft = f_fft+Yn1p_avg*static_max_freq*tot_matrix_normalization
        # sln_fft = jnp.linalg.solve(tot_matrix, f_fft)

        
        len_phi = Yn_rhs.content.shape[1]
        Yn_free_content = jnp.fft.ifft(fft_pad(sln_fft, len_phi, axis=0), axis=0)[None, :]
        # Yn_free_content = solve_ODE(
        #     coeff_arr=coef_Yn1c_in_D3.content,
        #     coeff_dp_arr=coef_dp_Yn1c_in_D3.content*nfp,
        #     f_arr=D3_RHS_no_unknown.content,
        #     static_max_freq=static_max_freq
        # ) # Seems underdetermined. Blows up.

    Yn = new_Y_n_no_unknown + ChiPhiFunc(vector_free_coef*Yn_free_content, nfp)
    return(Yn)

# Evaluates Zn,
# Requires:
# X_{n-1}, Y_{n-1}, Z_{n-1},
# B_{\theta n}, B_{\psi n-1},
# B_{\alpha (n-2)/2 \text{ or } (n-3)/2},
# \iota_{(n-2)/2 \text{ or } (n-3)/2}
# \kappa, \frac{dl}{d\phi}, \tau
# not nfp-dependent
def iterate_Zn_cp(
    n_eval,
    X_coef_cp, Y_coef_cp, Z_coef_cp,
    B_theta_coef_cp, B_psi_coef_cp,
    B_alpha_coef,
    kap_p, dl_p, tau_p,
    iota_coef):
    return(parsed.eval_Zn_cp(
        n=n_eval,
        X_coef_cp=X_coef_cp.mask(n_eval-1),
        Y_coef_cp=Y_coef_cp.mask(n_eval-1),
        Z_coef_cp=Z_coef_cp.mask(n_eval-1).zero_append(),
        # B_theta n isn't required but need to be provided (cancellation?)
        B_theta_coef_cp=B_theta_coef_cp.mask(n_eval-1).zero_append(),
        B_psi_coef_cp=B_psi_coef_cp.mask(n_eval-2),
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p, dl_p=dl_p, tau_p=tau_p,
        iota_coef = iota_coef
    ).cap_m(n_eval))

# Evaluates B_{\psi n-2}
# Requires:
# X_{n-1}, Y_{n-1}, Z_{n-1},
# B_{\theta n-1}, B_0,
# B_{\alpha 0}, \bar{\iota}_{(n-2)/2 or (n-3)/2}$
# not nfp-dependent
def iterate_dc_B_psi_nm2(
    n_eval,
    X_coef_cp, Y_coef_cp, Z_coef_cp,
    B_theta_coef_cp, B_psi_coef_cp,
    B_alpha_coef, B_denom_coef_c,
    kap_p, dl_p, tau_p,
    iota_coef):
    # Ignore B_theta_n
    dchi_b_psi_nm2 = parsed.eval_dchi_B_psi_cp_nm2(n_eval, \
            X_coef_cp.mask(n_eval-1).zero_append(), \
            Y_coef_cp.mask(n_eval-1).zero_append(), \
            Z_coef_cp.mask(n_eval-1).zero_append(), \
            B_theta_coef_cp.mask(n_eval), B_psi_coef_cp.mask(n_eval-3).zero_append(), B_alpha_coef, B_denom_coef_c, \
            kap_p, dl_p, tau_p, iota_coef)
    # The evaluation gives an extra component.
    if isinstance(dchi_b_psi_nm2, ChiPhiFunc):
        return(dchi_b_psi_nm2.cap_m(n_eval-2))
    else:
        return(dchi_b_psi_nm2)

''' II. Magnetic and MHD equations '''
''' II.1. Recursion relations for individual variables '''

# Uses B_theta [n-2], B_[psi n-2], B_alpha first-order terms, B_denom[n],
# p_[perp n-1], Delta_[n-1],iota_[(n-3)/2 (n-2)/2]
# not nfp-dependent
def iterate_p_perp_n(n_eval,
    B_theta_coef_cp,
    B_psi_coef_cp,
    B_alpha_coef,
    B_denom_coef_c,
    p_perp_coef_cp,
    Delta_coef_cp,
    iota_coef):
    return(
        MHD_parsed.eval_p_perp_n_cp(n_eval,
        B_theta_coef_cp.mask(n_eval-2).zero_append().zero_append(), # cancellation for 2 orders
        B_psi_coef_cp,
        B_alpha_coef,
        B_denom_coef_c,
        p_perp_coef_cp.mask(n_eval-1).zero_append(),
        Delta_coef_cp,
        iota_coef).cap_m(n_eval)
    )

# Uses B_denom[n], p_[\perp n]$, $\Delta_[n-1]$, $\iota_[(n-1)/2, n/2]$
# Delta always contain one free, scalar parameter that's determined through
# the looped equation. At even orders, this is avg(Delta[n,0]), the constant
# component of the FFT of Delta[n,0]. At odd orders, this is iota[(n-1)/2].
# This method always ensures zero avg(Delta[n,0]), and will by default
# set iota[(n-1)/2] to zero, but has the option to use known iota for debugging.
# nfp-dependent!!
def iterate_delta_n_0_offset(n_eval,
    B_denom_coef_c,
    p_perp_coef_cp,
    Delta_coef_cp,
    iota_coef,
    static_max_freq=-1,
    no_iota_masking = False): # nfp-dependent!!

    # At even orders, the free parameter is Delta_offset (the average of Delta n0)
    if n_eval%2==0:
        Delta_n_inhomog_component = MHD_parsed.eval_inhomogenous_Delta_n_cp(n_eval,
        B_denom_coef_c,
        p_perp_coef_cp,
        Delta_coef_cp.mask(n_eval-1).zero_append(),
        iota_coef)
    # At odd orders, the free parameter is iota (n-1)/2. Note the masking on iota_coef.
    # no_iota_masking is for debugging
    else:
        if no_iota_masking:
            Delta_n_inhomog_component = MHD_parsed.eval_inhomogenous_Delta_n_cp(n_eval,
            B_denom_coef_c,
            p_perp_coef_cp,
            Delta_coef_cp.mask(n_eval-1).zero_append(),
            iota_coef)
        else:
            Delta_n_inhomog_component = MHD_parsed.eval_inhomogenous_Delta_n_cp(n_eval,
            B_denom_coef_c,
            p_perp_coef_cp,
            Delta_coef_cp.mask(n_eval-1).zero_append(),
            iota_coef.mask((n_eval-3)//2).zero_append())

    # At even orders, setting Delta[even, 0] to have zero average.
    # This average is a free parameter, because the center component of
    # the ODE is dphi x = f.
    content = solve_dphi_iota_dchi(
        iota=iota_coef[0]/Delta_n_inhomog_component.nfp,
        f=Delta_n_inhomog_component.content/Delta_n_inhomog_component.nfp,
        static_max_freq=static_max_freq
    )
    Delta_out = ChiPhiFunc(content, Delta_n_inhomog_component.nfp).cap_m(n_eval)
    if n_eval%2==0:
        Delta_out -= jnp.average(Delta_out[0].content)
    return(Delta_out)

''' III. Equilibrium manager and Iterate '''

# A container for all equilibrium quantities.
# All coef inputs must be ChiPhiEpsFunc's.
class Equilibrium:
    # nfp-dependent!!
    def __init__(self, unknown, constant, axis_info, nfp, magnetic_only):
        self.unknown = unknown
        self.constant = constant
        self.nfp = nfp
        self.magnetic_only = magnetic_only
        self.axis_info = axis_info
        # Check if every term is on the same order
        # self.check_order_consistency()

    ''' For JAX use '''
    def _tree_flatten(self):
        children = (
            self.unknown,
            self.constant,
            self.axis_info,
        )  # arrays / dynamic values
        aux_data = {
            'nfp': self.nfp,
            'magnetic_only': self.magnetic_only
        }  # static values
        return (children, aux_data)
    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(*children, **aux_data)

    def from_known(
        X_coef_cp,
        Y_coef_cp,
        Z_coef_cp,
        B_psi_coef_cp,
        B_theta_coef_cp,
        B_denom_coef_c,
        B_alpha_coef,
        kap_p, dl_p, tau_p,
        iota_coef,
        p_perp_coef_cp,
        Delta_coef_cp,
        axis_info={},
        magnetic_only=False
        ):
        # Variables being solved for are stored in dicts for
        # convenience of plotting and saving
        unknown = {}
        constant = {}
        nfp = X_coef_cp.nfp
        unknown['X_coef_cp'] = X_coef_cp
        unknown['Y_coef_cp'] = Y_coef_cp
        unknown['Z_coef_cp'] = Z_coef_cp
        unknown['B_psi_coef_cp'] = B_psi_coef_cp
        unknown['B_theta_coef_cp'] = B_theta_coef_cp
        unknown['p_perp_coef_cp'] = p_perp_coef_cp
        unknown['Delta_coef_cp'] = Delta_coef_cp
        constant['iota_coef'] = iota_coef
        constant['B_denom_coef_c'] = B_denom_coef_c
        constant['B_alpha_coef'] = B_alpha_coef
        constant['kap_p'] = kap_p
        constant['dl_p'] = dl_p
        constant['tau_p'] = tau_p
        # Pressure can be trivial
        if not unknown['p_perp_coef_cp']:
            unknown['p_perp_coef_cp'] = ChiPhiEpsFunc.zeros_like(X_coef_cp)
        if not unknown['Delta_coef_cp']:
            unknown['Delta_coef_cp'] = ChiPhiEpsFunc.zeros_like(X_coef_cp)

        return(Equilibrium(
            unknown=unknown,
            constant=constant,
            axis_info=axis_info,
            nfp=nfp,
            magnetic_only=magnetic_only,
        ))
    
    ''' Coordinate transformations '''
    def frenet_basis_phi(self, phi):
        ''' 
        Calculates frenet basis vectors (in R, Phi, Z) from the flux coordinat phi. 
        '''
        varphi = self.axis_info['varphi']
        R0 = self.axis_info['R0']
        phi_grids = self.axis_info['phi']
        Z0 = self.axis_info['Z0']
        tangent_cylindrical = self.axis_info['tangent_cylindrical']
        normal_cylindrical = self.axis_info['normal_cylindrical']
        binormal_cylindrical = self.axis_info['binormal_cylindrical']
        nfp=self.nfp
        # axis location and basis in term of Boozer phi
         
        axis_r0_phi_R = jnp.interp(phi, varphi, R0, period = 2*jnp.pi/nfp)
        axis_r0_phi_Phi = jnp.interp(phi, varphi, phi_grids, period = 2*jnp.pi/nfp)
        axis_r0_phi_Z = jnp.interp(phi, varphi, Z0, period = 2*jnp.pi/nfp)
        tangent_phi_R = jnp.interp(phi, varphi, tangent_cylindrical[:, 0], period = 2*jnp.pi/nfp)
        tangent_phi_Phi = jnp.interp(phi, varphi, tangent_cylindrical[:, 1], period = 2*jnp.pi/nfp)
        tangent_phi_Z = jnp.interp(phi, varphi, tangent_cylindrical[:, 2], period = 2*jnp.pi/nfp)
        normal_phi_R = jnp.interp(phi, varphi, normal_cylindrical[:, 0], period = 2*jnp.pi/nfp)
        normal_phi_Phi = jnp.interp(phi, varphi, normal_cylindrical[:, 1], period = 2*jnp.pi/nfp)
        normal_phi_Z = jnp.interp(phi, varphi, normal_cylindrical[:, 2], period = 2*jnp.pi/nfp)
        binormal_phi_R = jnp.interp(phi, varphi, binormal_cylindrical[:, 0], period = 2*jnp.pi/nfp)
        binormal_phi_Phi = jnp.interp(phi, varphi, binormal_cylindrical[:, 1], period = 2*jnp.pi/nfp)
        binormal_phi_Z = jnp.interp(phi, varphi, binormal_cylindrical[:, 2], period = 2*jnp.pi/nfp)
        return(
            axis_r0_phi_R,
            axis_r0_phi_Phi,
            axis_r0_phi_Z,
            tangent_phi_R,
            tangent_phi_Phi,
            tangent_phi_Z,
            normal_phi_R,
            normal_phi_Phi,
            normal_phi_Z,
            binormal_phi_R,
            binormal_phi_Phi,
            binormal_phi_Z
        )

    def flux_to_frenet(self, psi, chi, phi, n_max=float('inf')):
        ''' 
        Transforms positions in the flux coordinate to the frenet frame.
        Returns (curvature, binormal, tangent) in the Frenet frame.
        '''
        if double_precision:
            target_type=jnp.float64
        else:
            target_type=jnp.float32
        return(
            self.unknown['X_coef_cp'].eval(psi=psi, chi=chi, phi=phi, n_max=n_max), 
            self.unknown['Y_coef_cp'].eval(psi=psi, chi=chi, phi=phi, n_max=n_max), 
            self.unknown['Z_coef_cp'].eval(psi=psi, chi=chi, phi=phi, n_max=n_max)
        )

    def flux_to_cylindrical(self, psi, chi, phi, n_max=float('inf')):
        ''' 
        Transforms positions in the flux coordinate to the 
        cylindrical coordinate.
        (R, Phi, Z) in the cylindrical coordinate.
        '''
        axis_r0_phi_R,\
        axis_r0_phi_Phi,\
        axis_r0_phi_Z,\
        tangent_phi_R,\
        tangent_phi_Phi,\
        tangent_phi_Z,\
        normal_phi_R,\
        normal_phi_Phi,\
        normal_phi_Z,\
        binormal_phi_R,\
        binormal_phi_Phi,\
        binormal_phi_Z = self.frenet_basis_phi(phi)
        curvature, binormal, tangent = self.flux_to_frenet(psi=psi, chi=chi, phi=phi, n_max=n_max)
        components_R = axis_r0_phi_R\
            + tangent_phi_R * tangent\
            + normal_phi_R * curvature\
            + binormal_phi_R * binormal
        
        components_Phi = axis_r0_phi_Phi\
            + tangent_phi_Phi * tangent\
            + normal_phi_Phi * curvature\
            + binormal_phi_Phi * binormal
        
        components_Z = axis_r0_phi_Z\
            + tangent_phi_Z * tangent\
            + normal_phi_Z * curvature\
            + binormal_phi_Z * binormal
        return(
            components_R,
            components_Phi,
            components_Z
        )

    def flux_to_xyz(self, psi, chi, phi, n_max=float('inf')):
        ''' 
        Transforms positions in the flux coordinate to the 
        XYZ coordinate.
        Returns (X, Y, Z) in the cylindrical coordinate.
        '''
        R, Phi, Z = self.flux_to_cylindrical(psi=psi, chi=chi, phi=phi, n_max=n_max)
        X = R*jnp.cos(Phi)
        Y = R*jnp.sin(Phi)
        return(X, Y, Z)

    ''' Saving and loading '''
    def save(self, file_name):
        jnp.save(file_name, self)

    def load(file_name):
        return(jnp.load(file_name, allow_pickle=True).item())

    def save_plain(self, file_name):
        ''' Saves to a npy file that can be read without aqsc or jax. '''
        unknown_dict = {}
        for key in self.unknown.keys():
            unknown_dict[key] = self.unknown[key].to_content_list()
        constant_dict={}
        constant_dict['B_denom_coef_c']\
            = self.constant['B_denom_coef_c'].to_content_list()
        constant_dict['B_alpha_coef']\
            = self.constant['B_alpha_coef'].to_content_list()
        constant_dict['iota_coef']\
            = self.constant['iota_coef'].to_content_list()
        constant_dict['kap_p']\
            = self.constant['kap_p'].content
        constant_dict['dl_p']\
            = self.constant['dl_p']
        constant_dict['tau_p']\
            = self.constant['tau_p'].content
        
        numpy_axis_info = {}
        for key in self.axis_info.keys():
            numpy_axis_info[key] = np.asarray(self.axis_info[key])

        big_dict = {
            'unknown':unknown_dict,
            'constant':constant_dict,
            'nfp':self.nfp,
            'magnetic_only':self.magnetic_only,
            'axis_info':numpy_axis_info,
        }
        np.save(file_name, big_dict)

    # nfp-dependent!!
    def load_plain(filename):
        npyfile = np.load(filename, allow_pickle=True)
        big_dict = npyfile.item()
        raw_unknown = big_dict['unknown']
        raw_constant = big_dict['constant']
        nfp = big_dict['nfp']
        magnetic_only = big_dict['magnetic_only']
        axis_info = big_dict['axis_info']
        print('nfp', nfp)

        unknown = {}
        for key in raw_unknown.keys():
            unknown[key] = ChiPhiEpsFunc.from_content_list(raw_unknown[key], nfp)

        constant={}
        constant['B_denom_coef_c']\
            = ChiPhiEpsFunc.from_content_list(raw_constant['B_denom_coef_c'], nfp)
        constant['B_alpha_coef']\
            = ChiPhiEpsFunc.from_content_list(raw_constant['B_alpha_coef'], nfp)
        constant['kap_p']\
            = ChiPhiFunc(raw_constant['kap_p'], nfp)
        constant['dl_p']\
            = raw_constant['dl_p']
        constant['tau_p']\
            = ChiPhiFunc(raw_constant['tau_p'], nfp)
        constant['iota_coef']\
            = ChiPhiFunc(raw_constant['iota_coef'], nfp)

        return(Equilibrium(
            unknown=unknown,
            constant=constant,
            axis_info=axis_info,
            nfp=nfp,
            magnetic_only=magnetic_only,
        ))

    # Order consistency check --------------------------------------------------
    # Get the current order of an equilibrium
    # not nfp-dependent
    def get_order(self):
        self.check_order_consistency()
        return(self.unknown['X_coef_cp'].get_order())

    # Check of all terms are on consistent orders
    # nfp=dependent!!
    def check_order_consistency(self):

        # An internal method for error throwing. Checks if all variables are
        # iterated to the correct order.
        def check_order_individial(ChiPhiEpsFunc, name_in_error_msg, n_req):
            n_current = ChiPhiEpsFunc.get_order()
            if n_current!=n_req:
                raise AttributeError(name_in_error_msg
                    + ' has order: '
                    + str(n_current)
                    + ', while required order is: '
                    + str(n_req))
            if ChiPhiEpsFunc.nfp!=0 and ChiPhiEpsFunc.nfp!=self.nfp:
                raise AttributeError(name_in_error_msg
                    + ' has nfp: '
                    + str(ChiPhiEpsFunc.nfp)
                    + ', while required nfp is: '
                    + str(self.nfp))


        n = self.unknown['X_coef_cp'].get_order()
        if n%2==1:
            raise AttributeError('X_coef_cp has odd order. Equilibrium is managed'\
            ' and updated every 2 orders and stops at even orders')

        check_order_individial(self.unknown['Y_coef_cp'], 'Y_coef_cp', n)
        check_order_individial(self.unknown['Z_coef_cp'], 'Z_coef_cp', n)
        check_order_individial(self.unknown['B_psi_coef_cp'], 'B_psi_coef_cp', n-2)
        check_order_individial(self.unknown['B_theta_coef_cp'], 'B_theta_coef_cp', n)
        check_order_individial(self.constant['B_denom_coef_c'], 'B_denom_coef_c', n)
        check_order_individial(self.constant['B_alpha_coef'], 'B_alpha_coef', n//2)
        check_order_individial(self.constant['iota_coef'], 'iota_coef', (n-2)//2)
        check_order_individial(self.unknown['p_perp_coef_cp'], 'p_perp_coef_cp', n)
        check_order_individial(self.unknown['Delta_coef_cp'], 'Delta_coef_cp', n)

    # Checks the accuracy of iteration at order n_unknown by substituting
    # results into the original form of the governing equations.
    # not nfp-dependent
    def check_governing_equations(self, n_unknown:int):
        # if n_unknown is None:
        #     n_unknown = self.get_order()
        # elif n_unknown>self.get_order():
        #     raise ValueError('Validation order should be <= than the current order')

        X_coef_cp = self.unknown['X_coef_cp']
        Y_coef_cp = self.unknown['Y_coef_cp']
        Z_coef_cp = self.unknown['Z_coef_cp']
        B_theta_coef_cp = self.unknown['B_theta_coef_cp']
        B_psi_coef_cp = self.unknown['B_psi_coef_cp']
        iota_coef = self.constant['iota_coef']
        p_perp_coef_cp = self.unknown['p_perp_coef_cp']
        Delta_coef_cp = self.unknown['Delta_coef_cp']
        B_denom_coef_c = self.constant['B_denom_coef_c']
        B_alpha_coef = self.constant['B_alpha_coef']
        kap_p = self.constant['kap_p']
        dl_p = self.constant['dl_p']
        tau_p = self.constant['tau_p']

        J = MHD_parsed.validate_J(n_unknown,
            X_coef_cp,
            Y_coef_cp,
            Z_coef_cp,
            B_denom_coef_c, B_alpha_coef,
            kap_p, dl_p, tau_p, iota_coef)
        Cb = MHD_parsed.validate_Cb(n_unknown-1,
            X_coef_cp,
            Y_coef_cp, Z_coef_cp,
            B_denom_coef_c, B_alpha_coef,
            B_psi_coef_cp, B_theta_coef_cp,
            kap_p, dl_p, tau_p, iota_coef)
        # The general expression does not work for the zeroth order, because it
        # performs order-matching with symbolic order n and then sub in n at
        # evaluation time, rather than actually performing order-matching at each
        # order. As a result, it IGNORES constant terms in the equation.
        if n_unknown==0:
            Cb-=dl_p
        Ck = MHD_parsed.validate_Ck(n_unknown-1, X_coef_cp, Y_coef_cp, Z_coef_cp,
            B_denom_coef_c, B_alpha_coef,
            B_psi_coef_cp, B_theta_coef_cp,
            kap_p, dl_p, tau_p, iota_coef)
        Ct = MHD_parsed.validate_Ct(n_unknown-1, X_coef_cp, Y_coef_cp, Z_coef_cp,
            B_denom_coef_c, B_alpha_coef,
            B_psi_coef_cp, B_theta_coef_cp,
            kap_p, dl_p, tau_p, iota_coef)
        if self.magnetic_only:
            I = ChiPhiFuncSpecial(0)
            II = ChiPhiFuncSpecial(0)
            III = ChiPhiFuncSpecial(0)
        else:
            I = MHD_parsed.validate_I(n_unknown, B_denom_coef_c,
                p_perp_coef_cp, Delta_coef_cp,
                iota_coef)
            II = MHD_parsed.validate_II(n_unknown,
                B_theta_coef_cp, B_alpha_coef, B_denom_coef_c,
                p_perp_coef_cp, Delta_coef_cp, iota_coef)
            III = MHD_parsed.validate_III(n_unknown-2,
            B_theta_coef_cp, B_psi_coef_cp,
            B_alpha_coef, B_denom_coef_c,
            p_perp_coef_cp, Delta_coef_cp,
            iota_coef)
        return(J, Cb, Ck, Ct, I, II, III)

    ''' Display and output'''

    def get_psi_crit(
        self, n_max=float('inf'), 
        n_grid_chi=100, n_grid_phi=100, 
        eps_cap = 3.0,
        n_newton_iter = 20):
        eps_crit, jacobian_residue = self.get_eps_crit(
            n_max=n_max, 
            n_grid_chi=n_grid_chi, 
            n_grid_phi=n_grid_phi, 
            eps_cap=eps_cap,
            n_newton_iter=n_newton_iter
        )
        return(eps_crit**2, jacobian_residue)
        

    def get_eps_crit(
        self, n_max=float('inf'), 
        n_grid_chi=100, n_grid_phi=100, 
        eps_cap = 2.0,
        n_newton_iter = 10):
        '''
        Estimates the critical epsilon where flux surface self-intersects.
        by finding the zero of $min_{\chi, \phi}[\sqrt{g}(\epsilon, \chi, \phi)]=0$
        using binary search. \sqrt{g}(\epsilon, \chi, \phi) At each search step 
        is evaluated on a grid of given size.
        Can be diffed but is slow to compile, because it relies on newton iteration
        for root finding. Reducing `n_newton_iter` substantially speeds up jit.

        Parameters;

        - `n_max` - maximum order to evaluate coordinate transformation to.

        - `n_grid_chi, n_grid_phi : int`- Grid size to evaluate Jacobian $\sqrt{g}$ on. 
        The critical point occurs when $min(\sqrt{g}\leq0)$.

        - `bad_eps_init` - An initial guess of an epsilon>epsilon_crit used in Newton's 
        method. Need to be beyond t

        - `n_newton_iter:int` (static) - Maximum number of steps in Newton's method.
        higher number gives better acuracy but is slower to jit.

        Returns: 
        
        - (eps_crit, jacobian_residue): eps_crit and the flux surface min of the Jacobian at eps_crit.
        '''
        points = jnp.linspace(0, 2*np.pi*(1-1/n_grid_phi), n_grid_phi)
        points_chi = jnp.linspace(0, 2*np.pi*(1-1/n_grid_chi), n_grid_chi)
        axis_r0_phi_R,\
        axis_r0_phi_Phi,\
        axis_r0_phi_Z,\
        tangent_phi_R,\
        tangent_phi_Phi,\
        tangent_phi_Z,\
        normal_phi_R,\
        normal_phi_Phi,\
        normal_phi_Z,\
        binormal_phi_R,\
        binormal_phi_Phi,\
        binormal_phi_Z = self.frenet_basis_phi(points)
        axis_r0 = jnp.real(jnp.array([
            axis_r0_phi_R,
            axis_r0_phi_Phi,
            axis_r0_phi_Z
        ]))
        tangent = jnp.real(jnp.array([
            tangent_phi_R,
            tangent_phi_Phi,
            tangent_phi_Z
        ]))
        normal = jnp.real(jnp.array([
            normal_phi_R,
            normal_phi_Phi,
            normal_phi_Z
        ]))
        binormal = jnp.real(jnp.array([
            binormal_phi_R,
            binormal_phi_Phi,
            binormal_phi_Z
        ]))

        # These quantities are periodic, and we apply
        # spectral derivatives, taking advantage of implementation
        # in aqsc.
        tangent_dphi = jnp.real(ChiPhiFunc(tangent, self.nfp).dphi().content)
        normal_dphi = jnp.real(ChiPhiFunc(normal, self.nfp).dphi().content)
        binormal_dphi = jnp.real(ChiPhiFunc(binormal, self.nfp).dphi().content)
        axis_r0_dphi = jnp.real(jnp.gradient(axis_r0, points[1]-points[0], axis=1)) # First and last 2 elems untrustworthy
        X_coef_cp = self.unknown['X_coef_cp']
        Y_coef_cp = self.unknown['Y_coef_cp']
        Z_coef_cp = self.unknown['Z_coef_cp']

        # Returns (n_grid, n_grid) arrays. Axis=1 is Phi.
        X_grid = lambda eps: jnp.real(X_coef_cp.eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        Y_grid = lambda eps: jnp.real(Y_coef_cp.eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        Z_grid = lambda eps: jnp.real(Z_coef_cp.eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        deps_X_grid = lambda eps: jnp.real(X_coef_cp.deps().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        deps_Y_grid = lambda eps: jnp.real(Y_coef_cp.deps().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        deps_Z_grid = lambda eps: jnp.real(Z_coef_cp.deps().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        dchi_X_grid = lambda eps: jnp.real(X_coef_cp.dchi().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        dchi_Y_grid = lambda eps: jnp.real(Y_coef_cp.dchi().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        dchi_Z_grid = lambda eps: jnp.real(Z_coef_cp.dchi().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        dphi_X_grid = lambda eps: jnp.real(X_coef_cp.dphi().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        dphi_Y_grid = lambda eps: jnp.real(Y_coef_cp.dphi().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))
        dphi_Z_grid = lambda eps: jnp.real(Z_coef_cp.dphi().eval(eps**2, points_chi[:, None], points[None, :], n_max=n_max))

        deps_r = lambda eps:(
            deps_X_grid(eps)[None, :, :]*normal[:, None, :]
            +deps_Y_grid(eps)[None, :, :]*binormal[:, None, :]
            +deps_Z_grid(eps)[None, :, :]*tangent[:, None, :]
        )

        dchi_r = lambda eps:(
            dchi_X_grid(eps)[None, :, :]*normal[:, None, :]
            +dchi_Y_grid(eps)[None, :, :]*binormal[:, None, :]
            +dchi_Z_grid(eps)[None, :, :]*tangent[:, None, :]
        )

        dphi_r = lambda eps:(
            axis_r0_dphi[:, None, :]
            +dphi_X_grid(eps)[None, :, :]*normal[:, None, :]
            +dphi_Y_grid(eps)[None, :, :]*binormal[:, None, :]
            +dphi_Z_grid(eps)[None, :, :]*tangent[:, None, :]
            +X_grid(eps)[None, :, :]*normal_dphi[:, None, :]
            +Y_grid(eps)[None, :, :]*binormal_dphi[:, None, :]
            +Z_grid(eps)[None, :, :]*tangent_dphi[:, None, :]
        )
        # Sqrt g on grid
        # HACKY: np.gradient seem unreliable for the leading and trailing 2 elements/
        # Because g is a smooth function, we remove these two cols and assume errors 
        # from neglecting 2 cols of grid points in argmin{min[g(r)]<=0} is negligible.
        jacobian_grid = lambda eps: jnp.sum(
            jnp.cross(
                deps_r(eps), 
                dchi_r(eps), 
                axisa=0, axisb=0, axisc=0
            )*dphi_r(eps), 
            axis=0
        )[:, 2:-2] 

        # Finding jacobian_min=0 using Newtom's method.
        jacobian_min = lambda eps: jnp.min(jacobian_grid(eps))
        jacobian_min_prime = jax.grad(jacobian_min)
        def q(x):
            return x - jacobian_min(x) / jacobian_min_prime(x)
        eps_i = eps_cap
        for i in range(n_newton_iter):
            y = q(eps_i)
            eps_i = y
        return(eps_i, jacobian_min(eps_i))
        '''
        # Zero-finding with binary search is faster but cannot be jitted.
        # Binary search for jacobian_min=0,
        # knowing jacobian_min(0)=0 and assuming jacobian_min is concave.
        x_left = 0
        x_right = bad_eps_init
        # g_tol is the tolerance and supposed to
        error = g_tol + 1
        n_iter = 0
        n_bindary_iter_max = 200
        # Tolerance-based stopping is disabled for now
        # while error > g_tol and n_iter < n_bindary_iter_max:
        # while error > g_tol and n_iter < n_bindary_iter_max:
            # We assume that jacobian_min is concave and is 0 at 0.
            x_mid = .5*x_left + .5*x_right
            y_mid = jacobian_min(x_mid)
            if y_mid>0:
                x_left = x_mid
            else: 
                x_right = x_mid
            error = jnp.abs(y_mid)
            n_iter += 1
        '''      



    def get_helicity(self):
        ''' 
        Returns the helicity (the normal vector, kappa's 
        # rotation around the origin)
        '''
        # R, phi, Z of the normal vector
        norm_R = np.pad(self.axis_info['normal_cylindrical'][:, 0], (0, 1), 'wrap')
        norm_Z = np.pad(self.axis_info['normal_cylindrical'][:, -1], (0, 1), 'wrap')
        norm_R = norm_R-np.average(norm_R) # zero-centering
        norm_Z = norm_Z-np.average(norm_Z) # zero-centering
        # To make arctan accumulate over or under +- pi,
        # we need to detect crosses between quadrant 2 and 3
        arctan_offset = np.zeros_like(norm_R)
        for i in range(len(norm_R)-1):
            # Sign change in Z
            if np.sign(norm_Z[i])*np.sign(norm_Z[i+1])==-1:
                Z_diff = norm_Z[i+1]-norm_Z[i]
                R_diff = norm_R[i+1]-norm_R[i]
                R_center = norm_R[i]+R_diff/Z_diff*(-norm_Z[i])
                # the line segment between two adjacent points 
                # crosses x<0, y=0.
                if R_center<0: 
                    # crossing from quadrant 2 to 3
                    if norm_Z[i]>0:
                        arctan_offset[i+1:]+=np.pi*2                
                    # crossing from quadrant 3 to 2
                    else:
                        arctan_offset[i+1:]-=np.pi*2
        arctan_accumulated = np.arctan2(norm_Z,norm_R)+arctan_offset
        return(round((arctan_accumulated[-1]-arctan_accumulated[0])/(2*np.pi)))

    def display(self, n_max=float('inf'), psi_max:float=0.2):
        fig = plt.figure()
        fig.set_dpi(400)
        ax = fig.add_subplot(projection='3d')
        phis = jnp.linspace(0, np.pi*2*0.9, 100)
        chis = jnp.linspace(0, np.pi*2, 100)
        x_surf, y_surf, z_surf = self.flux_to_xyz(psi=psi_max, chi=chis[None, :], phi=phis[:, None], n_max=n_max)
        # Coloring by magnitude of B
        B_denom = self.constant['B_denom_coef_c'].eval(
            psi=psi_max, chi=chis[None, :], phi=phis[:, None]
        )
        B_magnitude = 1/jnp.real(B_denom).astype(jnp.float32)

        norm = colors.Normalize(vmin=jnp.min(B_magnitude), vmax=jnp.max(B_magnitude), clip=True)
        mapper = cm.ScalarMappable(norm=norm, cmap=cm.plasma)
        facecolors = mapper.to_rgba(B_magnitude)

        ax.plot_surface(x_surf, y_surf, z_surf, zorder=1, facecolors=facecolors)
        ax.axis('equal')
        for psi_i in np.linspace(0, psi_max, 5):
            x_cross, y_cross, z_cross = self.flux_to_xyz(psi=psi_i, chi=chis, phi=0, n_max=n_max)
            ax.plot(x_cross, y_cross, z_cross, zorder=2.5, linewidth=0.5)
        fig.colorbar(mapper, label=r'$|B|^2$', shrink=0.5)
        fig.show()

    # not nfp-dependent
    def display_order(self, n:int):
        if n<2:
            raise ValueError('Can only display order n>1')
        for name in self.unknown.keys():
            if name!='B_psi_coef_cp':
                print(name, 'n =', n)
                n_display = n
            else:
                print(name, 'n =', n-2)
                n_display = n-2

            if type(self.unknown[name][n_display]) is ChiPhiFunc:
                self.unknown[name][n_display].display_content()
            else:
                print(self.unknown[name][n_display])

tree_util.register_pytree_node(
Equilibrium,
Equilibrium._tree_flatten,
Equilibrium._tree_unflatten)

''' Iteration '''
# # Evaluates 2 entire orders. Note that no masking is needed for any of the methods
# # defined in this file. Copies the equilibrium. STOPS AT EVEN ORDERS.

# Iterates the magnetic equations only.
# Calculates Xn, Yn, Zn, B_psin-2 for 2 orders from lower order values.
# B_theta, B_psi_nm30, Y_free_nm1 are all free.
# n_eval must be even.
# not nfp-dependent
# Yn0, B_psi_nm20 must both be consts or 1d arrays.
def iterate_2_magnetic_only(equilibrium,
    B_theta_nm1, B_theta_n,
    Yn0,
    Yn1c_avg,
    B_psi_nm20,
    B_alpha_nb2,
    B_denom_nm1, B_denom_n,
    iota_nm2b2,
    static_max_freq=(-1,-1),
    traced_max_freq=(-1,-1),
    n_eval=None,
):

    if not equilibrium.magnetic_only:
        return()
    if static_max_freq == None:
        len_phi = equilibrium.unknown['X_coef_cp'][1].content.shape[1]
        static_max_freq = (len_phi//2, len_phi//2)

    # If no order is supplied, then iterate to the next order. the equilibrium
    # will be edited directly.
    if n_eval == None:
        n_eval = equilibrium.get_order() + 2 # getting order and checking consistency
    if n_eval%2 != 0:
        raise ValueError("n must be even to evaluate iota_{(n-1)/2}")

    print("Evaluating order",n_eval-1, n_eval)

    # Creating new ChiPhiEpsFunc's for the resulting equilibrium
    X_coef_cp = equilibrium.unknown['X_coef_cp'].mask(n_eval-2)
    Y_coef_cp = equilibrium.unknown['Y_coef_cp'].mask(n_eval-2)
    Z_coef_cp = equilibrium.unknown['Z_coef_cp'].mask(n_eval-2)
    B_theta_coef_cp = equilibrium.unknown['B_theta_coef_cp'].mask(n_eval-2)
    B_psi_coef_cp = equilibrium.unknown['B_psi_coef_cp'].mask(n_eval-4)
    iota_coef = equilibrium.constant['iota_coef'].mask((n_eval-4)//2)
    p_perp_coef_cp = equilibrium.unknown['p_perp_coef_cp'].mask(n_eval-2)
    Delta_coef_cp = equilibrium.unknown['Delta_coef_cp'].mask(n_eval-2)

    # Masking all init conds.
    B_denom_coef_c = equilibrium.constant['B_denom_coef_c'].mask(n_eval-2)
    B_alpha_coef = equilibrium.constant['B_alpha_coef'].mask((n_eval)//2-1)
    kap_p = equilibrium.constant['kap_p']
    dl_p = equilibrium.constant['dl_p']
    tau_p = equilibrium.constant['tau_p']
    # Appending free functions and initial conditions
    B_theta_coef_cp = B_theta_coef_cp.append(B_theta_nm1)
    B_theta_coef_cp = B_theta_coef_cp.append(B_theta_n)
    B_alpha_coef = B_alpha_coef.append(B_alpha_nb2)
    iota_coef = iota_coef.append(iota_nm2b2)
    B_denom_coef_c = B_denom_coef_c.append(B_denom_nm1)
    B_denom_coef_c = B_denom_coef_c.append(B_denom_n)
    p_perp_coef_cp = p_perp_coef_cp.append(ChiPhiFuncSpecial(0))
    p_perp_coef_cp = p_perp_coef_cp.append(ChiPhiFuncSpecial(0))
    Delta_coef_cp = Delta_coef_cp.append(ChiPhiFuncSpecial(0))
    Delta_coef_cp = Delta_coef_cp.append(ChiPhiFuncSpecial(0))

    # Evaluating order n_eval-1
    # Requires:
    # X_{n-1}, Y_{n-1}, Z_{n-1},
    # B_{\theta n-1}, B_0,
    # B_{\alpha 0}, \bar{\iota}_{(n-2)/2 or (n-3)/2}$
    B_psi_nm3 = iterate_dc_B_psi_nm2(n_eval=n_eval-1,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        ).antid_chi()
    B_psi_coef_cp = B_psi_coef_cp.append(B_psi_nm3.filter(traced_max_freq[0]))

    # Requires:
    # X_{n-1}, Y_{n-1}, Z_{n-1},
    # B_{\theta n}, B_{\psi n-2},
    # B_{\alpha (n-2)/2 or (n-3)/2},
    # \iota_{(n-2)/2 or (n-3)/2}
    # \kappa, \frac{dl}{d\phi}, \tau
    Znm1 = iterate_Zn_cp(n_eval=n_eval-1,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        )
    Z_coef_cp = Z_coef_cp.append(Znm1.filter(traced_max_freq[0]))

    # Requires:
    # X_{n-1}, Y_{n-1}, Z_n,
    # \iota_{(n-3)/2 or (n-2)/2},
    # B_{\alpha (n-1)/2 or n/2}.
    Xnm1 = iterate_Xn_cp(n_eval=n_eval-1,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_denom_coef_c=B_denom_coef_c,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        )
    X_coef_cp = X_coef_cp.append(Xnm1.filter(traced_max_freq[0]))

    # Requires:
    # X_{n}, Y_{n-1}, Z_{n-1},
    # B_{\theta n-1}, B_{\psi n-3},
    # \iota_{(n-3)/2 or (n-4)/2}, B_{\alpha (n-1)/2 or (n-2)/2}
    Ynm1 = iterate_Yn_cp_magnetic(n_unknown=n_eval-1,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef,
        static_max_freq=static_max_freq[0],
        Yn1c_avg=Yn1c_avg
    )
    Y_coef_cp = Y_coef_cp.append(Ynm1.filter(traced_max_freq[0]))

    # Order n_eval ---------------------------------------------------
    # no need to ignore_mode_0 for chi integral. This is an odd order.
    B_psi_nm2 = iterate_dc_B_psi_nm2(n_eval=n_eval,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        ).antid_chi()
    B_psi_nm2_content_new = B_psi_nm2.content.at[B_psi_nm2.content.shape[0]//2].set(B_psi_nm20)
    B_psi_nm2 = ChiPhiFunc(B_psi_nm2_content_new, B_psi_nm2.nfp)
    B_psi_coef_cp = B_psi_coef_cp.append(B_psi_nm2.filter(traced_max_freq[1]))

    Zn = iterate_Zn_cp(n_eval=n_eval,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        )
    Z_coef_cp = Z_coef_cp.append(Zn.filter(traced_max_freq[1]))

    Xn = iterate_Xn_cp(n_eval=n_eval,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_denom_coef_c=B_denom_coef_c,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        )
    X_coef_cp = X_coef_cp.append(Xn.filter(traced_max_freq[1]))

    Yn = iterate_Yn_cp_magnetic(n_unknown=n_eval,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef,
        static_max_freq=static_max_freq[1],
        Yn0=Yn0
    )
    Y_coef_cp = Y_coef_cp.append(Yn.filter(traced_max_freq[1]))

    # return(X_coef_cp,
    #     Y_coef_cp,
    #     Z_coef_cp,
    #     B_psi_coef_cp,
    #     B_theta_coef_cp,
    #     B_denom_coef_c,
    #     B_alpha_coef,
    #     kap_p,
    #     dl_p,
    #     tau_p,
    #     iota_coef,
    #     p_perp_coef_cp,
    #     Delta_coef_cp,
    # )

    return(Equilibrium.from_known(
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_denom_coef_c=B_denom_coef_c,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef,
        p_perp_coef_cp=p_perp_coef_cp,
        Delta_coef_cp=Delta_coef_cp,
        magnetic_only=True,
        axis_info=equilibrium.axis_info
    ))

# @partial(jit, static_argnums=(5, ))
def iterate_2(equilibrium,
    B_alpha_nb2,
    B_denom_nm1, B_denom_n,
    # Not implemented yet. At odd orders, the periodic BC
    # will constrain one of sigma_n(0), iota_n and avg(B_theta_n0)
    # given value for the other 2.
    # free_param_values need to be a 2-element tuple.
    # Now only implemented avg(B_theta_n0)=0 and given iota.
    iota_new, # arg 4
    static_max_freq=(-1,-1),
    traced_max_freq=(-1,-1),
    # Traced.
    # -1 represents no filtering (default). This value is chosen so that
    # turning on or off off-diagonal filtering does not require recompiles.
    max_k_diff_pre_inv=(-1,-1),
    n_eval=None,
    ):
    if equilibrium.magnetic_only:
        return()
    if n_eval == None:
        n_eval = equilibrium.get_order() + 2 # getting order and checking consistency
    if n_eval%2 != 0:
        raise ValueError("n must be even to evaluate iota_{(n-1)/2}")

    print("Evaluating order",n_eval-1, n_eval)

    # Resetting unknowns
    X_coef_cp = equilibrium.unknown['X_coef_cp'].mask(n_eval-2)
    Y_coef_cp = equilibrium.unknown['Y_coef_cp'].mask(n_eval-2)
    Z_coef_cp = equilibrium.unknown['Z_coef_cp'].mask(n_eval-2)
    B_theta_coef_cp = equilibrium.unknown['B_theta_coef_cp'].mask(n_eval-2)
    B_psi_coef_cp = equilibrium.unknown['B_psi_coef_cp'].mask(n_eval-4)
    iota_coef = equilibrium.constant['iota_coef'].mask((n_eval-4)//2)
    p_perp_coef_cp = equilibrium.unknown['p_perp_coef_cp'].mask(n_eval-2)
    Delta_coef_cp = equilibrium.unknown['Delta_coef_cp'].mask(n_eval-2)

    # Masking all init conds.
    B_denom_coef_c = equilibrium.constant['B_denom_coef_c'].mask(n_eval-2)
    B_alpha_coef = equilibrium.constant['B_alpha_coef'].mask((n_eval)//2-1)
    kap_p = equilibrium.constant['kap_p']
    dl_p = equilibrium.constant['dl_p']
    tau_p = equilibrium.constant['tau_p']

    iota_coef = iota_coef.append(iota_new)
    B_denom_coef_c = B_denom_coef_c.append(B_denom_nm1)
    B_denom_coef_c = B_denom_coef_c.append(B_denom_n)
    B_alpha_coef = B_alpha_coef.append(B_alpha_nb2)

    # Evaluating order n_eval-1

    # print('iota 1 right before loop',iota_coef[1])
    solution_nm1_known_iota = looped_solver.iterate_looped(
        n_unknown=n_eval-1,
        nfp=equilibrium.nfp,
        target_len_phi=1000,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        p_perp_coef_cp=p_perp_coef_cp,
        Delta_coef_cp=Delta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        tau_p=tau_p,
        dl_p=dl_p,
        iota_coef=iota_coef,
        static_max_freq=static_max_freq[0],
        traced_max_freq=traced_max_freq[0],
        max_k_diff_pre_inv=max_k_diff_pre_inv[0],
    )
    B_theta_coef_cp = B_theta_coef_cp.append(solution_nm1_known_iota['B_theta_n']) 
    B_psi_coef_cp = B_psi_coef_cp.append(solution_nm1_known_iota['B_psi_nm2']) 
    X_coef_cp = X_coef_cp.append(solution_nm1_known_iota['Xn']) 
    Y_coef_cp = Y_coef_cp.append(solution_nm1_known_iota['Yn']) 
    Z_coef_cp = Z_coef_cp.append(solution_nm1_known_iota['Zn']) 
    p_perp_coef_cp = p_perp_coef_cp.append(solution_nm1_known_iota['pn']) 
    Delta_coef_cp = Delta_coef_cp.append(solution_nm1_known_iota['Deltan']) 

    # This "partial" solution will be fed into
    # iterate_looped. This is already filtered.
    B_theta_coef_cp = B_theta_coef_cp.append(solution_nm1_known_iota['B_theta_np10'])

    # B_psi[n-2] without the zeroth component.
    # MAY NOT BE NECESSARY
    B_psi_nm2 = iterate_dc_B_psi_nm2(n_eval=n_eval,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef
        ).antid_chi()
    B_psi_coef_cp = B_psi_coef_cp.append(B_psi_nm2)

    solution_n = looped_solver.iterate_looped(
        n_unknown=n_eval,
        nfp=equilibrium.nfp,
        target_len_phi=1000,
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        p_perp_coef_cp=p_perp_coef_cp,
        Delta_coef_cp=Delta_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_alpha_coef=B_alpha_coef,
        B_denom_coef_c=B_denom_coef_c,
        kap_p=kap_p,
        tau_p=tau_p,
        dl_p=dl_p,
        iota_coef=iota_coef,
        static_max_freq=static_max_freq[1],
        traced_max_freq=traced_max_freq[1],
        max_k_diff_pre_inv=max_k_diff_pre_inv[1],
    )
    # Partial solutions for these variables were appended to their
    # ChiPhiEpsFunc's for iterate_looped. Now remove them and re-append
    # This only reassigns the pointer B_psi. Need to re-assign equilibrium.unknown[]
    # too.
    B_psi_coef_cp = B_psi_coef_cp.mask(n_eval-3)
    B_psi_coef_cp = B_psi_coef_cp.append(solution_n['B_psi_nm2'].filter(traced_max_freq[1]))
    # This only reassigns the pointer B_theta. Need to re-assign equilibrium.unknown[]
    # too.
    B_theta_coef_cp = B_theta_coef_cp.mask(n_eval-1)
    B_theta_coef_cp = B_theta_coef_cp.append(solution_n['B_theta_n'].filter(traced_max_freq[1]))

    X_coef_cp = X_coef_cp.append(solution_n['Xn']) 
    Z_coef_cp = Z_coef_cp.append(solution_n['Zn']) 
    p_perp_coef_cp = p_perp_coef_cp.append(solution_n['pn']) 
    Delta_coef_cp = Delta_coef_cp.append(solution_n['Deltan']) 
    Y_coef_cp = Y_coef_cp.append(solution_n['Yn']) 

    return(Equilibrium.from_known(
        X_coef_cp=X_coef_cp,
        Y_coef_cp=Y_coef_cp,
        Z_coef_cp=Z_coef_cp,
        B_psi_coef_cp=B_psi_coef_cp,
        B_theta_coef_cp=B_theta_coef_cp,
        B_denom_coef_c=B_denom_coef_c,
        B_alpha_coef=B_alpha_coef,
        kap_p=kap_p,
        dl_p=dl_p,
        tau_p=tau_p,
        iota_coef=iota_coef,
        p_perp_coef_cp=p_perp_coef_cp,
        Delta_coef_cp=Delta_coef_cp,
        magnetic_only=False,
        axis_info=equilibrium.axis_info
    ))



import unittest
from aqsc import *
import pathlib

import jax.numpy as jnp

# The numerical derivatives usually aren't as accurate 
# as to go below the threshold of np.isclose().
print('Testing all parsed recursion relations '\
      '(except the looped equation) against '\
      'Rodriguez\'s circular axis calculation')


circ_axis_tolerance = 1e-4
def is_roughly_close_chiphifunc(chiphifunc_a, chiphifunc_b):
    return(
        jnp.all(
            jnp.abs(chiphifunc_a.content-chiphifunc_b.content)<circ_axis_tolerance
        )
    )

test_dir = str(pathlib.Path(__file__).parent.resolve())
B_psi_coef_cp, B_theta_coef_cp, \
    Delta_coef_cp, p_perp_coef_cp,\
    X_coef_cp, Y_coef_cp, Z_coef_cp, \
    iota_coef, dl_p,\
    nfp, Xi_0, eta, \
    B_denom_coef_c, B_alpha_coef, \
    kap_p, tau_p = read_first_three_orders(
        test_dir+'/circ/', 
        R_array=[2,0,1,2,0.0001,0],
        Z_array=[1,2,0,0.001]
    )

equilibrium = Equilibrium.from_known(
    X_coef_cp=X_coef_cp.mask(2),
    Y_coef_cp=Y_coef_cp.mask(2),
    Z_coef_cp=Z_coef_cp.mask(2),
    B_psi_coef_cp=B_psi_coef_cp.mask(0),
    B_theta_coef_cp=B_theta_coef_cp.mask(2),
    B_denom_coef_c=B_denom_coef_c.mask(2),
    B_alpha_coef=B_alpha_coef.mask(1),
    iota_coef=iota_coef.mask(0), 
    kap_p=kap_p, 
    dl_p=dl_p, 
    tau_p=tau_p,
    p_perp_coef_cp = p_perp_coef_cp.mask(2), 
    Delta_coef_cp = Delta_coef_cp.mask(2))

class TestCircularAxis(unittest.TestCase):

    def test_B_psi(self):
        print('Testing B_psi iteration' )
        B_psi_nm2 = iterate_dc_B_psi_nm2(n_eval=3,
            X_coef_cp=X_coef_cp,
            Y_coef_cp=Y_coef_cp,
            Z_coef_cp=Z_coef_cp,
            B_theta_coef_cp=B_theta_coef_cp.zero_append(),
            B_psi_coef_cp=B_psi_coef_cp,
            B_alpha_coef=B_alpha_coef,
            B_denom_coef_c=B_denom_coef_c,
            kap_p=kap_p,
            dl_p=dl_p,
            tau_p=tau_p,
            iota_coef=iota_coef
            ).antid_chi()
        print_fractional_error(B_psi_nm2.content,B_psi_coef_cp[1].content)
        self.assertTrue(is_roughly_close_chiphifunc(B_psi_nm2, B_psi_coef_cp[1]))

    def test_X(self):
        print('Testing X iteration' )
        X2 = iterate_Xn_cp(2,
            X_coef_cp,
            Y_coef_cp,
            Z_coef_cp,
            B_denom_coef_c,
            B_alpha_coef,
            kap_p, dl_p, tau_p,
            iota_coef)
        print_fractional_error(X2.content, X_coef_cp[2].content)
        self.assertTrue(is_roughly_close_chiphifunc(X2, X_coef_cp[2]))

    def test_Z(self):
        print('Testing Z iteration' )
        Z2 = iterate_Zn_cp(
            2,
            X_coef_cp, Y_coef_cp, Z_coef_cp,
            B_theta_coef_cp, B_psi_coef_cp,
            B_alpha_coef,
            kap_p, dl_p, tau_p,
            iota_coef)
        print_fractional_error(Z2.content, Z_coef_cp[2].content)
        self.assertTrue(is_roughly_close_chiphifunc(Z2, Z_coef_cp[2]))

    def test_Y(self):
        print('Testing Y iteration' )
        Yn = iterate_Yn_cp_magnetic(
            2,
            X_coef_cp,
            Y_coef_cp,
            Z_coef_cp,
            B_psi_coef_cp,
            B_theta_coef_cp,
            B_alpha_coef,
            B_denom_coef_c,
            kap_p, dl_p, tau_p,
            iota_coef,
            static_max_freq=50, 
            Yn0=Y_coef_cp[2][0].content)
        print_fractional_error(Yn.content, Y_coef_cp[2].content)
        self.assertTrue(is_roughly_close_chiphifunc(Yn, Y_coef_cp[2]))

    def test_p(self):
        print('Testing p iteration' )
        pn = iterate_p_perp_n(2,
            B_theta_coef_cp,
            B_psi_coef_cp,
            B_alpha_coef,
            B_denom_coef_c,
            p_perp_coef_cp,
            Delta_coef_cp,
            iota_coef)
        print_fractional_error(pn.content, p_perp_coef_cp[2].content)
        self.assertTrue(is_roughly_close_chiphifunc(pn, p_perp_coef_cp[2]))

    def test_Delta(self):
        print('Testing Delta iteration' )
        Delta_n = iterate_delta_n_0_offset(2,
            B_denom_coef_c,
            p_perp_coef_cp,
            Delta_coef_cp,
            iota_coef,
            static_max_freq=None,
            no_iota_masking = False)
        # Delta_n[0]'s average need to be solved at the following order. 
        # Here we compare two Deltas after setting both of their m=0
        # components to have zero average.
        Delta_n_no_avg = Delta_n - jnp.average(Delta_n[0].content)
        Delta_ans_no_avg = Delta_coef_cp[2] - jnp.average(Delta_coef_cp[2][0].content)
        print_fractional_error(Delta_n_no_avg.content, Delta_ans_no_avg.content)
        self.assertTrue(is_roughly_close_chiphifunc(Delta_n_no_avg, Delta_ans_no_avg))
        
if __name__ == '__main__':
    unittest.main()


from aqsc.math_utilities import *

def eval_y_coefs(n_eval, X_coef_cp, Y_coef_cp,
    Delta_coef_cp, B_alpha_coef, dl_p, tau_p, iota_coef):
    coef_Y = (
        2*(Delta_coef_cp[0]-1)*(diff(X_coef_cp[1],True,1))*dl_p*(diff(tau_p,False,1))
        +(
            2*(Delta_coef_cp[0]-1)*iota_coef[0]*(diff(X_coef_cp[1],True,2))
            +2*((Delta_coef_cp[0]-1)*diff(X_coef_cp[1],True,1)).dphi()
        )*dl_p*tau_p
        +(1-Delta_coef_cp[0])*iota_coef[0]**2*(diff(Y_coef_cp[1],True,3))
        +2*(1-Delta_coef_cp[0])*iota_coef[0]*(diff(Y_coef_cp[1],True,2,False,1))
        -iota_coef[0]*(diff(Delta_coef_cp[0],False,1))*(diff(Y_coef_cp[1],True,2))
        +(1-Delta_coef_cp[0])*(diff(Y_coef_cp[1],True,1,False,2))
        -(diff(Delta_coef_cp[0],False,1))*(diff(Y_coef_cp[1],True,1,False,1))
    )/B_alpha_coef[0]*(n_eval-1)/n_eval

    coef_dchi_Y = -(
        (2*Delta_coef_cp[0]-2)*X_coef_cp[1]*dl_p*(diff(tau_p,False,1))
        +(
            (2*Delta_coef_cp[0]-2)*(diff(X_coef_cp[1],False,1))
            +2*(n_eval-2)*(1-Delta_coef_cp[0])*iota_coef[0]*(diff(X_coef_cp[1],True,1))
            +2*X_coef_cp[1]*(diff(Delta_coef_cp[0],False,1))
        )*dl_p*tau_p
        +(1-Delta_coef_cp[0])*(diff(Y_coef_cp[1],False,2))
        -(diff(Delta_coef_cp[0],False,1))*(diff(Y_coef_cp[1],False,1))
        +(1-Delta_coef_cp[0])*iota_coef[0]**2*(diff(Y_coef_cp[1],True,2))
        +(2-2*Delta_coef_cp[0])*iota_coef[0]*(diff(Y_coef_cp[1],True,1,False,1))
        -n_eval*iota_coef[0]*(diff(Delta_coef_cp[0],False,1))*(diff(Y_coef_cp[1],True,1))
    )/B_alpha_coef[0]/n_eval

    coef_dphi_Y = (
        2*(Delta_coef_cp[0]-1)*(diff(X_coef_cp[1],True,1))*dl_p*tau_p
        +(diff(Delta_coef_cp[0],False,1))*(diff(Y_coef_cp[1],True,1))
    )/B_alpha_coef[0]*(n_eval-1)/n_eval

    coef_dchi_dphi_Y = -(
        (2*Delta_coef_cp[0]-2)*X_coef_cp[1]*dl_p*tau_p
        +(n_eval-1)*(2-2*Delta_coef_cp[0])*iota_coef[0]*(diff(Y_coef_cp[1],True,1))
        +Y_coef_cp[1]*(diff(Delta_coef_cp[0],False,1))
    )/B_alpha_coef[0]/n_eval

    coef_dphi_dphi_Y = (
        (Delta_coef_cp[0]-1)*diff(Y_coef_cp[1],True,1)
    )/B_alpha_coef[0]*(n_eval-1)/n_eval

    coef_dchi_dchi_Y = -(
        (2*Delta_coef_cp[0]-2)*iota_coef[0]*X_coef_cp[1]*dl_p*tau_p
        +(n_eval-1)*(1-Delta_coef_cp[0])*iota_coef[0]**2*(diff(Y_coef_cp[1],True,1))
        +iota_coef[0]*Y_coef_cp[1]*(diff(Delta_coef_cp[0],False,1))
    )/B_alpha_coef[0]/n_eval

    coef_dchi_dchi_dphi_Y = -(2*Delta_coef_cp[0]-2)*iota_coef[0]*Y_coef_cp[1]/B_alpha_coef[0]/n_eval

    coef_dphi_dphi_dchi_Y = -(Delta_coef_cp[0]-1)*Y_coef_cp[1]/B_alpha_coef[0]/n_eval

    coef_dchi_dchi_dchi_Y = -(Delta_coef_cp[0]-1)*iota_coef[0]**2*Y_coef_cp[1]/B_alpha_coef[0]/n_eval

    return({
        'coef_Y':coef_Y.cap_m(1),
        'coef_dchi_Y':coef_dchi_Y.cap_m(1),
        'coef_dphi_Y':coef_dphi_Y.cap_m(1),
        'coef_dchi_dphi_Y':coef_dchi_dphi_Y.cap_m(1),
        'coef_dphi_dphi_Y':coef_dphi_dphi_Y.cap_m(1),
        'coef_dchi_dchi_Y':coef_dchi_dchi_Y.cap_m(1),
        'coef_dchi_dchi_dphi_Y':coef_dchi_dchi_dphi_Y.cap_m(1),
        'coef_dphi_dphi_dchi_Y':coef_dphi_dphi_dchi_Y.cap_m(1),
        'coef_dchi_dchi_dchi_Y':coef_dchi_dchi_dchi_Y.cap_m(1),
    })



''' Where global settings are defined '''
# APES can run without pyQSC, but can have additional features
#  when pyQSC is available. Setting to True enables:
# chiphifunc.py
# - 'pseudo_spectral' mode for ChiPhiFunc.dphi()
# chiphifunc_test_suite.py
# - import_from_stel(): Loading equilibria from pyQSC
# - read_first_three_orders(): Loading equilibria from ref [1-3]
use_pyQSC = False

''' Numerical settings (chiphifunc.py) '''
double_precision = True
diff_mode = 1 # 1 for fft, 2 for pseudo_spectral
# Currently, It takes ~1hr to compile iterate_2 (most of the time taken by
# looped_solver.py) to order 4 for each combination of nfp and number of
# harmonics. Compiling to higher order may require ~10 hrs.
# However, it offers ~ 20x speed up for each case.
# The compile can be disabled to only compile recursion relations for individual
# variables and save
compile_MHD_iteration = False

''' Caching and Joblib settings '''
# Caching -----
# Maximum size of LRU cache in ChiPhiFunc
max_size = 1000

''' Plotting and output settings (for chiphifunc_test_suite.py) '''
# Number of grid points when:
# 1. Generating random ChiPhiFunc and ChiPhiEpsFunc for testing
# using rand_ChiPhiFunc() rand_ChiPhiEpsFunc()
# 2. Evaluating the values represented by power series or power series
# coefficients using evaluate(lambda), evaluate_ChiPhiFunc(ChiPhiFunc) and
# evaluate_ChiPhiEpsFunc(ChiPhiEpsFunc)
n_grid_phi = 1000
n_grid_chi = 500



import unittest
from aqsc import *
import jax.numpy as jnp
import numpy as np

# Size of the chi and phi grid used for evaluation tests
points = jnp.linspace(0, 2*np.pi*(1-1/n_grid_phi), n_grid_phi)
chi = points
phi = points
psi = jnp.linspace(0,5,100)

# Generating test objects
content_single_nfp = jnp.array([
    jnp.sin(4*points), # sin component
    jnp.cos(4*points) # cos component
])
content1 = jnp.array([
    jnp.sin(points), # sin component
    jnp.cos(points) # cos component
])
single_period = ChiPhiFunc(content_single_nfp, 1, trig_mode=True)
four_period = ChiPhiFunc(content1, nfp=4, trig_mode=True)

content0 = jnp.array([
    jnp.sin(9*points)
])
content1 = jnp.array([
    jnp.sin(3*points), # sin component
    jnp.cos(4*points) # cos component
])
content2 = jnp.array([
    jnp.cos(6*points), # sin component
    jnp.sin(3*points), # 0 component
    jnp.cos(4*points) # cos component
])
content3 = jnp.array([
    jnp.sin(3*points), # sin component
    jnp.cos(4*points), # sin component
    jnp.sin(3*points), # cos component
    jnp.cos(4*points) # cos component
])
chiphifunc0 = ChiPhiFunc(content0, nfp=4, trig_mode=True)
chiphifunc1 = ChiPhiFunc(content1, nfp=4, trig_mode=True)
chiphifunc2 = ChiPhiFunc(content2, nfp=4, trig_mode=True)
chiphifunc3 = ChiPhiFunc(content3, nfp=4, trig_mode=True)
test_chiphiepsfunc = ChiPhiEpsFunc([
    chiphifunc0, 
    chiphifunc1, 
    chiphifunc2, 
    chiphifunc3
], nfp=4)


# np.gradient is not very accurate, but it allows us to benchmark our
# different implementation
solver_tolerance = 5e-4
def is_roughly_close(array_a, array_b):
    return(jnp.abs(array_a-array_b)<solver_tolerance)

class TestCallables(unittest.TestCase):
    

    def test_chiphifunc_eval(self):
        '''
        Testing function evaluation.
        '''
        lambda_guess = four_period.eval(chi[:, None], phi[None, :])
        lambda_guess2 = single_period.eval(chi[:, None], phi[None, :])
        lambda_ans = jnp.sin(chi[:, None])*jnp.sin(4*points) + jnp.cos(chi[:, None])*jnp.cos(4*points)
        self.assertTrue(jnp.all(jnp.isclose(lambda_guess, lambda_ans)))
        self.assertTrue(jnp.all(jnp.isclose(lambda_guess2, lambda_ans)))

    def test_chiphiepsfunc_square_series_lambda(self):
        '''
        Testing lambda function output for power series like:
        Sum(iota[n]*eps**(2*i), 0, n)
        = Sum(iota[n]*psi**i, 0, n)
        '''
        test_iota = ChiPhiEpsFunc([1.2, 3.6, 0, 5.2], 1)
        lambda_psi_ans = lambda psi: 1.2+3.6*psi+5.2*psi**3

        self.assertTrue(jnp.all(jnp.isclose(
            test_iota.eval(psi, sq_eps_series=True), 
            lambda_psi_ans(psi)
        )))
    
    def test_chiphiepsfunc_lambda(self):
        def lambda_ans(psi, chi, phi):
            phi = phi*4
            eps = jnp.sqrt(psi)
            out = jnp.sin(9*phi)
            
            out += jnp.sin(3*phi)*jnp.sin(chi)*eps
            out += jnp.cos(4*phi)*jnp.cos(chi)*eps
            
            out += jnp.cos(6*phi)*jnp.sin(2*chi)*eps**2
            out += jnp.sin(3*phi)*eps**2
            out += jnp.cos(4*phi)*jnp.cos(2*chi)*eps**2

            out += jnp.sin(3*phi)*jnp.sin(3*chi)*eps**3
            out += jnp.cos(4*phi)*jnp.sin(1*chi)*eps**3
            out += jnp.sin(3*phi)*jnp.cos(1*chi)*eps**3
            out += jnp.cos(4*phi)*jnp.cos(3*chi)*eps**3

            return(out)

        self.assertTrue(jnp.all(jnp.isclose(
            test_chiphiepsfunc.eval(psi[:,None,None], chi[None,:,None], phi[None,None,:]),
            lambda_ans(psi[:,None,None], chi[None,:,None], phi[None,None,:])
        )))

    def test_chiphiepsfunc_diff(self):
        epsilons = jnp.linspace(0,10,1000)
        phis = jnp.linspace(0,2*jnp.pi*0.999,1000)
        X_epsilons = test_chiphiepsfunc.eval(epsilons**2, 0, 0)
        X_chis = test_chiphiepsfunc.eval(1, phis, 0)
        X_phis = test_chiphiepsfunc.eval(1, 0, phis)
        gradient_eps_X = np.gradient(X_epsilons, epsilons)
        gradient_chi_X = np.gradient(X_chis, phis)
        gradient_phi_X = np.gradient(X_phis, phis)
        # First and last elements of jnp.gradient are incorrect.
        jnp.all(is_roughly_close(test_chiphiepsfunc.deps().eval(epsilons**2, 0, 0), gradient_eps_X)[1:-1])
        jnp.all(is_roughly_close(test_chiphiepsfunc.dchi().eval(1, phis, 0), gradient_chi_X)[1:-1])
        jnp.all(is_roughly_close(test_chiphiepsfunc.dphi().eval(1, 0, phis), gradient_phi_X)[1:-1]) # First and last elements of np.gradient are unreliable


if __name__ == '__main__':
    unittest.main()